(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_job_exec.ml 239 2010-06-23 16:49:03Z gerd $ *)

(** Where to run which task:
   - map task: is specified in the task
   - sort task: same node as preceding map
   - first round of shuffling: same node as preceding sort
   - further rounds of shuffling: any node
   - reduce task: same node as preceding shuffling task if this task
     produced only one file; otherwise any node
 *)

open Mapred_tasks
open Mapred_def
open Printf

module Mr_proxy = Mapred_rpc_clnt.Make'Mapred_task(Rpc_proxy.ManagedClient)


type runtime_job_config =
    { map_weight : float * float;
      sort_weight : float * float;
      shuffle_weight : float * float;
      reduce_weight : float * float;
      io_load : float;
      cpu_load : float;
      dump_plan_when_complete : bool;
      shm_low : int64;
      shm_high : int64;
    }

let create_runtime_job_config 
      ?(map_weight=(0.9, 0.1)) 
      ?(sort_weight=(0.1, 0.9)) 
      ?(shuffle_weight=(0.9, 0.1))
      ?(reduce_weight=(0.5, 0.5)) 
      ?(dump_plan_when_complete=false)
      ?(shm_low = Int64.max_int)
      ?(shm_high = Int64.max_int)
      ~io_load ~cpu_load () =
  { map_weight = map_weight;
    sort_weight = sort_weight;
    shuffle_weight = shuffle_weight;
    reduce_weight = reduce_weight;
    io_load = io_load;
    cpu_load = cpu_load;
    dump_plan_when_complete = dump_plan_when_complete;
    shm_low = shm_low;
    shm_high = shm_high;
  }

module MCSet = Set.Make(Rpc_proxy.ManagedClient)

type status =
    [ `Running | `Successful | `Errors of string list | `Killed ]

type running_job =
    { config : runtime_job_config;
      plan : Mapred_sched.plan;
      mutable started_tasks : task list;  (* for cleanup *)
      mutable killed : bool;
      mutable running : (int, task * Unix.inet_addr) Hashtbl.t;
        (* key: request id *)
      mutable clients : MCSet.t;
      mutable status : status;
      esys : Unixqueue.event_system;
      port : int;
      work_dir : string;
    }


let misalloc_penalty = 4.0
  (** Misallocated tasks have this factor bigger weight. The idea is to
     prevent misallocations as long as good allocations are possible.
     (A misallocated task is run on a non-preferred host.)
   *)

let new_client job saddr =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Mapred_rpc_clnt.Mapred_task.V1._program ]
      ~initial_ping:true
      () in
  (* TODO: no timeout so far - this is not possible with current RPC API *)
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      (Plasma_util.connector_of_sockaddr saddr)
      job.esys in
  job.clients <- MCSet.add mclient job.clients;
  Mr_proxy.V1.configure'async 
    mclient 
    (job.config.shm_low, job.config.shm_high)
    (fun _ -> ());
  mclient



let dispose_client job c =
  job.clients <- MCSet.remove c job.clients


let kill job =
  Unix.sleep 1;  (* give servers time to log errors *)
  Netlog.logf `Info "Killing job";
  job.killed <- true;
  Hashtbl.iter
    (fun req_id (task,ip) ->
       let task_s = Mapred_tasks.string_of_task_id task in
       Netlog.logf `Info "Submitting kill request of %s to %s"
	 task_s (Unix.string_of_inet_addr ip);
       let saddr = Unix.ADDR_INET(ip,job.port) in
       let c = new_client job saddr in
       Mr_proxy.V1.kill'async
	 c
	 req_id
	 (fun get_reply ->
	    Rpc_proxy.ManagedClient.shut_down c;
	    try
	      let () = get_reply() in
	      Netlog.logf `Info "Kill request for %s was accepted" task_s
	    with
	      | error ->
		  Netlog.logf `Err
		    "Exception after kill request for %s to %s: %s"
		    task_s (Unix.string_of_inet_addr ip)
		    (Netexn.to_string error)
	 )
    )
    job.running;
  Hashtbl.clear job.running;
  if job.status = `Running then
    job.status <- `Killed
    


let start esys plan mj rjc args port =
  let load = Hashtbl.create 5 in
  (** (io_load,cpu_load) per IP addr *)

  let plan_dumped = ref false in

  let job =
    { config = rjc;
      plan = plan;
      started_tasks = [];
      killed = false;
      running = Hashtbl.create 7;
      clients = MCSet.empty;
      status = `Running;
      esys = esys;
      port = port;
      work_dir = mj#work_dir
    } in


  let get_load ip =
    try Hashtbl.find load ip with Not_found -> (0.0, 0.0) in

  let next_req_id = ref 0 in

  let submit task pref_hosts (io_weight,cpu_weight) onsuccess =
    (** Try to start [task] on the [pref_hosts]. If there are no preferences,
       tries to start on any host. Returns [true] if possible, else [false].
     *)
    let hosts =
      if pref_hosts = [] then 
	List.map snd (Mapred_sched.hosts plan)
      else
	pref_hosts in
    if hosts = [] then
      failwith "List of hosts is empty";
    let hosts_sorted =
      List.sort
	(fun ip1 ip2 ->
	   let (io_ld1, cpu_ld1) = get_load ip1 in
	   let (io_ld2, cpu_ld2) = get_load ip2 in
	   compare (io_ld1 +. cpu_ld1) (io_ld2 +. cpu_ld2)
	)
	hosts in
    let ip_opt =
      try
	Some
	  (List.find
	     (fun ip ->
		let (ip_io_load, ip_cpu_load) = get_load ip in
		ip_io_load +. io_weight <= rjc.io_load && 
		  ip_cpu_load +. cpu_weight <= rjc.cpu_load
	     )
	     hosts_sorted)
      with Not_found -> None in
    match ip_opt with
      | Some ip ->
	  let (ip_io_load, ip_cpu_load) = get_load ip in
	  let task_s = Mapred_tasks.string_of_task_id task in
	  let saddr = Unix.ADDR_INET(ip,port) in
	  let c = new_client job saddr in
	  let req_id = !next_req_id in
	  incr next_req_id;
	  Mr_proxy.V1.exec'async
	    c
	    (req_id, args, Mapred_tasks.encode_task task)
	    (fun get_reply ->
	       let res =
		 try Mapred_tasks.decode_task_result(get_reply())
		 with error ->
		   Netlog.logf `Err "Exception from exec to %s: %s"
		     (Unix.string_of_inet_addr ip)
		     (Netexn.to_string error);
		   `Error("RPC error: " ^ Netexn.to_string error) in
	       Rpc_proxy.ManagedClient.shut_down c;
	       dispose_client job c;
	       let (old_ip_io_load, old_ip_cpu_load) = get_load ip in
	       let new_ip_io_load = old_ip_io_load -. io_weight in
	       let new_ip_cpu_load = old_ip_cpu_load -. cpu_weight in
	       Hashtbl.replace load ip (new_ip_io_load, new_ip_cpu_load);
	       Hashtbl.remove job.running req_id;
	       match res with
		 | `Ok files ->
		     Netlog.logf `Info "Finished task %s" task_s;
		     ( match task with
			 | `Map mt ->
			     Mapred_sched.add_map_output
			       plan mj mt files
			 | _ -> ()
		     );
		     if rjc.dump_plan_when_complete && 
		         Mapred_sched.plan_complete plan && not !plan_dumped
		     then (
		       printf "COMPLETE PLAN:\n";
		       Mapred_sched.print_plan plan;
		       flush stdout;
		       plan_dumped := true
		     );
		     Mapred_sched.mark_as_finished plan task;
		     onsuccess()
		 | `Error msg ->
		     Netlog.logf `Err "Error for task %s: %s" task_s msg;
		     let x_msg = sprintf "Task %s: %s" task_s msg in
		     job.status <- `Errors [ x_msg ];
		     kill job
	    );
	  let new_ip_io_load = ip_io_load +. io_weight in
	  let new_ip_cpu_load = ip_cpu_load +. cpu_weight in
	  Hashtbl.replace load ip (new_ip_io_load, new_ip_cpu_load);
	  Hashtbl.replace job.running req_id (task,ip);
	  Mapred_sched.mark_as_started plan task;
	  job.started_tasks <- task :: job.started_tasks;
	  Netlog.logf `Info
	    "Submitted task %s to %s"
	    task_s
	    (Unix.string_of_inet_addr ip);
	  true
      | None ->
	  false
  in

  let rec determine_pref_hosts task =
    (** see comment above *)
    match task with
      | `Map mt -> 
	  mt.map_best_hosts
      | `Sort st ->
	  List.flatten
	    (List.map
	       determine_pref_hosts
	       (Mapred_sched.task_depends_on_list plan task))
      | `Shuffle st ->
	  List.flatten
	    (List.map
	       determine_pref_hosts
	       (Mapred_sched.task_depends_on_list plan task))
  in

  let determine_weight task misalloc_flag =
    let factor = 
      if misalloc_flag then misalloc_penalty else 1.0 in
    let (w_io, w_cpu) =
      match task with
	| `Map _ -> rjc.map_weight
	| `Sort _ -> rjc.sort_weight
	| `Shuffle t -> 
	    if t.shuffle_reduce then
	      rjc.reduce_weight
	    else
	      rjc.shuffle_weight in
    (factor *. w_io, w_cpu)
  in

  let rec loop () =
    let tasks = Mapred_sched.executable_tasks plan in
    Netlog.logf `Info
      "Stats: runnable=%d running=%d finished=%d total=%d complete=%B"
      (List.length tasks)
      (Mapred_sched.n_running plan)
      (Mapred_sched.n_finished plan)
      (Mapred_sched.n_total plan)
      (Mapred_sched.plan_complete plan);
    if not job.killed && tasks <> [] then (
      (** Get the pref_hosts for each task. Also, split into tasks with
	  preference and in tasks without. The tasks with pref will be
	  tried first for getting them allocated
       *)
      let tasks_prefs =
	List.map
	  (fun t ->
	     (t, determine_pref_hosts t)
	  )
	  tasks in
      let tasks_with, tasks_without =
	List.partition
	  (fun (t, prefs) -> prefs <> [])
	  tasks_prefs in
      (* TODO (perf): in the following, avoid to check for further job 
	 submissions if all nodes are already maxed out
       *)
      (** First try to start tasks with preferred hosts. If we reach
	  the load limit, tasks may not be startable, so some rest
	  will be left
       *)
      let tasks_with_nonstart =
	List.filter
	  (fun (t, prefs) ->
	     let weight = determine_weight t false in
	     let is_started = submit t prefs weight loop in
	     not is_started
	  )
	  tasks_with in
      (** Now try to start any other tasks *)
      List.iter
	(fun (t, _) ->
	   let weight = determine_weight t false in
	   ignore(submit t [] weight loop)
	)
	tasks_without;
      List.iter
	(fun (t, _) ->
	   let weight = determine_weight t true in
	   ignore(submit t [] weight loop)
	)
	tasks_with_nonstart;
    )
    else
      if Mapred_sched.plan_finished plan then (
	Netlog.logf `Info "Finished job";
	job.status <- `Successful
      )
  in

  Netlog.logf `Info "Starting job";
  loop();
  job


let status job = job.status


let files_to_keep task =
  (** This only looks at *_del flags *)
  match task with
    | `Map mt ->
	[]
    | `Sort st ->
	if not st.sort_input_del then
	  List.map (fun (n,_,_) -> n) st.sort_input
	else
	  []
    | `Shuffle st ->
	if not st.shuffle_input_del then
	  List.flatten
	    (List.map 
	       (fun (f,_,_) -> 
		  List.map (fun (n,_,_) -> n) f
	       )
	       st.shuffle_input
	    )
	else
	  []

let cleanup job =
  Netlog.logf `Info "Cleaning up";
  let to_keep = Hashtbl.create 5 in
  List.iter
    (fun task ->
       let keep = files_to_keep task in
       List.iter (fun f -> Hashtbl.replace to_keep f ()) keep;
    )
    job.started_tasks;
  let c = Mapred_sched.cluster job.plan in
  let trans = Plasma_client.start c in
  let to_delete = Plasma_client.list trans job.work_dir in
  Plasma_client.commit trans;

  List.iter
    (fun (filebase,_) ->
       let file = job.work_dir ^ "/" ^ filebase in
       if not (Hashtbl.mem to_keep file) then (
	 let trans = Plasma_client.start c in
	 ignore(Plasma_client.unlink trans file);
	 Plasma_client.commit trans
       )
    )
    to_delete;
  Netlog.logf `Info "Done clean-up"
