(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_elect.ml 239 2010-06-23 16:49:03Z gerd $ *)

open Plasma_util.Operators
open Printf

module Elect_proxy = Pfs_rpcapi_clnt.Make'Elect(Rpc_proxy.ManagedClient)

type role =
    [ `Coordinator | `Slave | `Excluded | `Unknown ]

module Role_var =
  Netplex_cenv.Make_var_type(struct type t = role end)

module String_var =
  Netplex_cenv.Make_var_type(struct type t = string end)

(** For simplicity we use global variables to store the current state.
    The election is only started once, and there is absolutely no
    concurrency.
 *)

let role = ref (`Unknown : role)
  (** The role. This is also started in a shared variable [rolename] *)

let rolename = "current_role"

let revision = ref None
  (** The revision from the db *)

let coordname = "current_coordinator"
  (** The coordinator *)

let cancel_announcements = ref (fun () -> ())
  (** This function will be initialized after starting the elect engine,
      so that calling it cancels the announcements. This is done when
      the [set_coordinator] RPC of this process is called.
   *)


let get_role() = 
  (** If the role is [`Unknown] we have to look into the shared variable.
      Otherwise just use our local copy (values other than [`Unknown] never
      change again)
   *)
  if !role = `Unknown then (
    let r =
      try
	Role_var.get rolename
      with _ -> `Unknown    (** the shared var is not yet initialized *)
    in
    role := r;
    r
  )
  else
    !role


let set_role r =
  role := r;
  Role_var.set rolename r


let get_coordinator() =
  try
    String_var.get coordname
  with
    | _ ->
	failwith "Nn_elect.get_coordinator"

let get_revision () =
  match !revision with
    | None ->
	let esys = Unixqueue.create_unix_event_system() in
	let db = Nn_db.transact esys in
	let e = Nn_db.revision_get_e db#connection esys in
	Uq_engines.when_state
	  ~is_done:(fun _ -> ignore(db#commit()))
	  ~is_error:(fun _ -> db#return())
	  ~is_aborted:(fun _ -> db#return())
	  e;
	Unixqueue.run esys;
	( match e#state with
	    | `Done r ->
		Netlog.logf `Info
		  "Election: Using local revision %s" r;
		revision := Some r;
		r
	    | `Error e ->
		Netlog.logf `Crit
		  "Election: Exception %s" (Netexn.to_string e);
		raise e
	    | _ ->
		assert false
	)
    | Some r ->
	r


let rec send_announcements_stream_e count conn announcement esys =
  (** Send the announcement to the remote server endlessly until a response
      comes back
   *)
  let rcache_config =
    Rpc_proxy.ReliabilityCache.create_rcache_config
      ~policy:`None () in
  let rcache =
    Rpc_proxy.ReliabilityCache.create_rcache rcache_config in
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Pfs_rpcapi_clnt.Elect.V1._program ]
      ~msg_timeout:1.0
      ~msg_timeout_is_fatal:false
      ~initial_ping:false
      ~exception_handler:(fun err -> 
			    Netlog.logf `Warning
			      "non-critical RPC exception: %s"
			      (Netexn.to_string err))
      ~rcache
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      esys in
  ( Plasma_util.rpc_engine
      mclient
      Elect_proxy.V1.announce'async
      announcement
    >> (function
	  | `Done ok -> 
	      `Done(Some ok)
	  | `Error err -> 
	      `Done None
	  | `Aborted ->
	      `Done None
       )
  )
  ++ (fun ok_opt ->
	Rpc_proxy.ManagedClient.shut_down mclient;
	match ok_opt with
	  | None ->
	      (** Just repeat. We introduce here a delay to avoid
		  a busy wait
		   *)
	      Plasma_util.delay_engine
		1.0 
		(fun () ->
		   (** Note that this recursion is problematic when
                       run too often - after all we are wrapping engines
                       in engines endlessly
		    *)
		   send_announcements_stream_e 
		     (count+1) conn announcement esys)
		esys
	  | Some ok ->
	      (** We have a final result *)
	      eps_e (`Done ok) esys
     )


let send_annocunements nn_config port esys =
  (** port: namenode port of this server *)
  let announcement =
    { Pfs_rpcapi_aux.ann_clustername = nn_config#nn_clustername;
      ann_sender = sprintf "%s:%d" nn_config#nn_hostname port;
      ann_eligible = Array.of_list nn_config#nn_nodes;
      ann_revision = get_revision();
      ann_rank = nn_config # nn_rank;
    } in
  let nodes =
    List.map Plasma_util.parse_host_port nn_config#nn_nodes in
  let dest_nodes =
    List.filter
      (fun (h,p) ->
	 h <> nn_config#nn_hostname || p <> port
      )
      nodes in
  (** Start sending announcements to all neighbor namenodes. *)
  let replying_nodes = ref [] in
  let all_ok = ref true in
  let canceled = ref false in
  let e =
    Plasma_util.msync_engine
      (List.map
	 (fun (h,p) ->
	    let conn = 
	      Plasma_util.connector_of_sockaddr
		(Plasma_util.sockaddr_of_host_port (h,p)) in
	    send_announcements_stream_e 0 conn announcement esys
	    >> (function
		  | `Done ok ->
		      Netlog.logf `Info
			"Election: Namenode %s:%d %s"
			h p
			(if ok then "approves" else "objects");
		      all_ok := !all_ok && ok;
		      replying_nodes := (h,p) :: !replying_nodes;
		      `Done ()
		  | (`Error _|`Aborted) as other ->
		      other
	       )
	 )
	 dest_nodes
      )
      esys in
  (** We cancel [e] after some time - only nodes that responded until then
      will be considered. It is also possible that [e] is stopped because
      somebody called [cancel_announcements]
   *)
  let g = Unixqueue.new_group esys in
  Unixqueue.once esys g nn_config#nn_elect_timeout
    (fun () -> e#abort() );
  cancel_announcements :=
    (fun () -> canceled := true; e#abort() );
  (** If [e] is finished (either fully successful, or aborted), and there
      was no cancellation, we return all nodes that responded and are
      alive.
   *)
  e >>
    (function
       | `Done () 
       | `Aborted ->
	   if not !canceled && !all_ok then
	     `Done (Some !replying_nodes)
	   else
	     `Done None
       | `Error e ->
	   Netlog.logf `Err
	     "Nn_elect.send_announcements: Exception %s"
	     (Netexn.to_string e);
	   `Error e
    )


let set_coordinator_at nn_config rev port (h,p) esys =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Pfs_rpcapi_clnt.Elect.V1._program ]
      ~msg_timeout:60.0
      ~msg_timeout_is_fatal:true
      ~initial_ping:false
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      (Plasma_util.connector_of_sockaddr
	 (Plasma_util.sockaddr_of_host_port (h,p)))
      esys in
  Plasma_util.rpc_engine
    mclient
    Elect_proxy.V1.set_coordinator'async
    ( sprintf "%s:%d" nn_config#nn_hostname port, nn_config#nn_clustername,rev )
  >> (fun st ->
	Rpc_proxy.ManagedClient.shut_down mclient;
	st
     )


let set_coordinator_everywhere_e nn_config rev port nodes esys =
  Plasma_util.msync_engine
    (List.map
       (fun (h,p) ->
	  set_coordinator_at nn_config rev port (h,p) esys
       )
       nodes
    )
    esys


let elect_e nn_config port esys =
  (** First send out announcement. If we get to the point when the result
      is available, and it is positive for this process, this process becomes
      the coordinator.
   *)
  let rev = get_revision() in
  send_annocunements nn_config port esys
  ++ (fun nodes_opt ->
	match nodes_opt with
	  | None ->
	      Netlog.logf `Info
		"Election: coordinator role was rejected by other nodes";
	      eps_e (`Done(false,rev)) esys
	  | Some nodes ->
	      Netlog.logf `Info
		"Election: got responses from %d node(s)" (List.length nodes);
	      if List.length nodes + 1 >= nn_config # nn_node_alive_min_startup
	      then (
		Netlog.logf `Info
		  "Election: becoming coordinator";
		set_role `Coordinator;
		set_coordinator_everywhere_e nn_config rev port nodes esys
		>> (function
		      | `Done () -> `Done (true,rev)
		      | `Aborted -> `Aborted
		      | `Error e -> `Error e
		   )
	      )
	      else (
		Netlog.logf `Alert
		  "Election: Not enough live namenodes. Not starting up";
		eps_e (`Done (false,rev)) esys
	      )
     )


let proc_announce nn_config ann =
  let my_rev = get_revision() in
  let my_rank = nn_config # nn_rank in
  let ann_rev = ann.Pfs_rpcapi_aux.ann_revision in
  let ann_rank = ann.Pfs_rpcapi_aux.ann_rank in
  let ann_clustername = ann.Pfs_rpcapi_aux.ann_clustername in
  
  if !role = `Unknown && nn_config#nn_clustername = ann_clustername then
    ann_rev > my_rev || 
      (ann_rev = my_rev && ann_rank < my_rank)
  else
    false


let proc_set_coordinator nn_config (coord_hp, cname, rev) =
  if nn_config#nn_clustername <> cname then (
    Netlog.logf `Err
      "Election: set_coordinator was called with wrong clustername";
    failwith "proc_set_coordinator: wrong clustername";
  );
  let my_rev = get_revision() in
  !cancel_announcements();
  match !role with
    | `Unknown ->
	Netlog.logf `Info
	  "Election: Coordinator node is %s (revision %s)"
	  coord_hp
	  rev;
	String_var.set coordname coord_hp;
	if rev = my_rev then (
	  Netlog.logf `Info "Election: Becoming slave node";
	  set_role `Slave
	)
	else (
	  Netlog.logf `Alert "Election: This node is excluded from operation \
                              because the revision is out of date";
	  set_role `Excluded
	)
    | _ ->
	Netlog.logf `Alert
	  "Election: Got a second set_coordinator call. This is likely to be a \
           serious error!"
	  


let setup_elect_api srv nn_config =
  Pfs_rpcapi_srv.Elect.V1.bind
    ~proc_null:(fun () -> ())
    ~proc_announce:(proc_announce nn_config)
    ~proc_set_coordinator:(proc_set_coordinator nn_config)
    srv
