(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_state.ml 239 2010-06-23 16:49:03Z gerd $ *)

open Printf
open Nn_config
open Plasma_util.Operators

module A = Pfs_rpcapi_aux

type errno =
    Plasma_util.errno

exception Transaction_error of errno

let transerr_e code esys =
  eps_e (`Error(Transaction_error code)) esys

let () =
  Netexn.register_printer
    (Transaction_error `enotrans)
    (fun e ->
       match e with
	 | Transaction_error te ->
	     let te_text = Plasma_util.string_of_errno te in
	     "Transaction_error(" ^ te_text ^ ")"
	 | _ -> assert false
    )

module StrSet = Set.Make(String)

class type shared_state_t =
object
  method event_system : Unixqueue.event_system
  method node_config : nn_node_config
  method cur_rev_ident : string
  method new_rev_ident : unit -> string
  method set_rev_ident : string -> unit
  method datastore_max_id : int option
  method datastore_use_id : int -> unit
  method inode_max_id : int64 option
  method inode_use_id : int64 -> unit
  method inode_get_new_id_serialized :
           (Unixqueue.event_system -> int64 Uq_engines.engine) ->
           int64 Uq_engines.engine
  method finish_transaction_serialized :
           (Unixqueue.event_system -> bool Uq_engines.engine) ->
           bool Uq_engines.engine
  method slaves : (string * Nn_push.client * bool ref) list
  method enough_slaves : bool
  method master : string
  method master_enabled : bool ref
  method exists_blockmap : string -> bool
  method create_blockmap : Nn_blockmap.blockmap_t -> unit
  method blockmaps : Nn_blockmap.blockmap_t list
  method get_blockmap : string -> Nn_blockmap.blockmap_t
  method blockmap_get_e : identity:string -> 
                              Nn_blockmap.blockmap_t Uq_engines.engine
  method blockmap_prioritizer : unit Uq_engines.prioritizer
  method forget_identity : string -> unit
  method lock_inode : int64 -> < > -> (unit -> unit) option
  method unlock_inodes : < > -> unit
  method lock_filename : string -> [`Link | `Unlink | `Exists] -> < > -> 
                         (unit -> unit) option
  method unlock_filenames : < > -> unit
end


type rpolicy =
    [ `Balanced of int ]

type journal =
    Nn_db.modification Queue.t

type transacted_datastores =
    { td_idlist : int list;
      td_safetrans_id : int64;
      td_safetrans_secret : int64;
      td_safetrans_tmo : int64;
    }


class type view_t =
object
  method pre_commit : journal
  method post_commit : bool -> unit
  method rollback : unit -> unit
end


class type ds_view_t =
object
  inherit view_t
  method datastore_new_id_e : unit -> int Uq_engines.engine
  method datastore_list_e : unit -> 
                              Nn_datastores.datastore list Uq_engines.engine
  method datastore_upd_e : id:int ->
                           identity:string ->
                           size:int64 ->
                           enabled:bool ->
                              unit Uq_engines.engine
  method datastore_del_e : id:int ->
                             unit Uq_engines.engine
end


class type bm_view_t =
object
  inherit view_t
  method blockmap_get_e : identity:string ->
                            Nn_blockmap.blockmap_t Uq_engines.engine
  method blockmap_alloc_e : rpolicy:rpolicy ->
                            index:int64 ->
                            number:int ->
                            pref:string list ->
                            transacted_datastores ->
                              Pfs_rpcapi_aux.blocklist Uq_engines.engine
  method blockmap_dealloc_e : (string * int64) list ->
                                unit Uq_engines.engine
  method blockmap_load_e : unit -> unit Uq_engines.engine
  method blockmap_keep : (string * int64) list -> unit
end


class type inode_view_t =
object
  inherit view_t
  method inode_get_e : id:int64 ->
                         Pfs_rpcapi_aux.inodeinfo option Uq_engines.engine
  method inode_alloc_e : Pfs_rpcapi_aux.inodeinfo ->
                            int64 Uq_engines.engine
  method inode_upd_e : id:int64 ->
                       keep_blocklimit:bool -> 
                       Pfs_rpcapi_aux.inodeinfo ->
                         unit Uq_engines.engine
  method inode_dealloc_e :  id:int64 -> unit Uq_engines.engine
  method inode_get_blocks_e : id:int64 ->
                              blkidx:int64 ->
                              len:int64 ->
                              bm_view_t ->
                                Pfs_rpcapi_aux.blockinfo list Uq_engines.engine 
  method inode_upd_blocks_e : id:int64 ->
                              del:(int64 * int64) ->
                              add:Pfs_rpcapi_aux.blockinfo list ->
                              set_mtime:bool ->
                              set_blocklimit:int64 ->
                                unit Uq_engines.engine
  method inode_upd_time : id:int64 ->
                          mtime:Pfs_rpcapi_aux.time option ->
                          ctime:Pfs_rpcapi_aux.time option ->
                              unit
  method filename_get_e : name:string -> 
                             int64 option Uq_engines.engine
  method filename_rev_get_e : inode:int64 ->
                                string list Uq_engines.engine
  method filename_count_e : inode:int64 ->
                              int Uq_engines.engine
  method filename_list_e : name:string ->
                               (string * int64) list Uq_engines.engine
  method filename_link_e : name:string -> parent:string -> inode:int64 ->
                             int64 Uq_engines.engine
  method filename_unlink_e : name:string ->
                               int64 Uq_engines.engine

  method delayed_inode_deletes : int64 list
    
end

let encap_time t =
  let t1 = Int64.of_float t in
  let t2 = truncate ((t -. Int64.to_float t1) *. 1E9) in
  { A.tsecs = t1; A.tnsecs = t2 }


(**********************************************************************)
(* Shared state                                                       *)
(**********************************************************************)

let create_slave_mclient esys timeout hostport =
  let (h,p) = Plasma_util.parse_host_port hostport in
  let addr = Plasma_util.sockaddr_of_host_port (h,p) in  (** May perform DNS lookup *)
  let conn = Plasma_util.connector_of_sockaddr addr in
  Nn_push.create_client conn timeout esys
  

class shared_state esys (conf:nn_node_config) master_index init_rev 
                   : shared_state_t =
  let master = List.nth conf#nn_nodes master_index in
object(self)
  val mutable rev_t = 0L
  val mutable rev_cnt = 0

  method event_system = esys

  method node_config = conf

  val mutable cur_rev_ident = init_rev

  method cur_rev_ident = cur_rev_ident

  method new_rev_ident() =
    (** Create unique identifiers. We take the time, and count the number
        of identifiers we already created per second. There is also a 
        random number part to avoid that other instances of the namenode
        can generate the same identifiers
     *)
    let t = Int64.of_float (Unix.time()) in
    if t = rev_t then (
      assert(rev_cnt < 999_999_999);
      rev_cnt <- rev_cnt+1;
    )
    else
      rev_cnt <- 0;
    rev_t <- t;
    let r =
      sprintf "%011Ld.%09d-%s"
	t rev_cnt (Digest.to_hex (Plasma_rng.random_bytes 16)) in
    r

  method set_rev_ident r =
    cur_rev_ident <- r;



  val mutable datastore_max_id = None

  method datastore_max_id = datastore_max_id

  method datastore_use_id id =
    let overwrite =
      match datastore_max_id with
	| None -> true
	| Some m -> id > m in
    if overwrite then
      datastore_max_id <- Some id

  val mutable inode_max_id = None

  method inode_max_id = inode_max_id

  method inode_use_id id =
    let overwrite =
      match inode_max_id with
	| None -> true
	| Some m -> id > m in
    if overwrite then
      inode_max_id <- Some id

  val inode_get_new_id_serializer = new Uq_engines.serializer esys

  method inode_get_new_id_serialized :
     (Unixqueue.event_system -> int64 Uq_engines.engine) ->
     int64 Uq_engines.engine =
    fun f ->
      inode_get_new_id_serializer # serialized f

  val finish_transaction_serializer = new Uq_engines.serializer esys

  method finish_transaction_serialized :
     (Unixqueue.event_system -> bool Uq_engines.engine) ->
     bool Uq_engines.engine =
    fun f ->
      finish_transaction_serializer # serialized f

(*
  val finish_db_conn_lz = lazy(Nn_db.transact esys)
   
  method finish_db_conn = Lazy.force finish_db_conn_lz
 *)

  val slaves = 
    List.rev
      (snd
	 (List.fold_left
	    (fun (i,l) hp ->
	       if i = master_index then
		 (i+1,l)
	       else
		 let slave = 
		   create_slave_mclient esys conf#nn_node_timeout hp in
		 let node =
		   (hp, slave, ref true) in   (* whether alive and enabled *)
		 (i+1,node::l)
	    ) 
	    (0,[]) 
	    conf#nn_nodes
	 )
      )

  val master_enabled = ref true

  method master = master

  method slaves = slaves

  method private slave_alive_count =
    List.fold_left
      (fun acc (_,_,enabled) -> if !enabled then acc+1 else acc)
      0
      slaves

  method enough_slaves =
    self#slave_alive_count + 1 >= conf#nn_node_alive_min
     (* +1: add one because the master is also a node... *)

  method master_enabled = master_enabled


  val blockmaps = Hashtbl.create 19 
  val blockmap_caches = Hashtbl.create 19 

  method exists_blockmap id =
    Hashtbl.mem blockmaps id

  method create_blockmap bm =
    Netlog.logf `Debug
      "Creating blockmap for %s"
      bm#identity;
    Hashtbl.replace blockmaps bm#identity bm

  method blockmaps =
    Hashtbl.fold (fun _ bm l -> bm::l) blockmaps []

  method get_blockmap identity =
    Hashtbl.find blockmaps identity

  method blockmap_get_e ~identity =
    try
      let bm = Hashtbl.find blockmaps identity in
      new Uq_engines.epsilon_engine (`Done bm) esys
    with
      | Not_found ->
	  Netlog.logf `Debug
	    "Missing blockmap for %s - looking into cache" identity;
	  ( try
	      let bm_cache = Hashtbl.find blockmap_caches identity in
	      bm_cache#get_engine()
	    with
	      | Not_found ->
		  Netlog.logf `Debug
		    "Missing cache for %s - creating cache and loading"
		    identity;
		  let bm_cache =
		    Uq_engines.cache
		      (fun esys ->
			 let ds = Nn_datastore_news.get ~wait:true identity in
			 let id = ds.Nn_datastores.id in
			 let size = ds.Nn_datastores.size in
			 Nn_db.with_ro_trans
			   (Nn_blockmap.blockmap_from_db_e
			      ~id ~identity ~size)
			 ++ (fun v ->
			       self # create_blockmap v;
			       eps_e (`Done v) esys
			    )
		      )
		      esys in
		  Hashtbl.add blockmap_caches identity bm_cache;
		  bm_cache#get_engine()
	  )

  method blockmap_prioritizer =
    Uq_engines.prioritizer esys

  method forget_identity id =
    (** Deleting/disabling datastores is a fragile operation. Remember that we
        might have handed out blockmap objects, or that the blockmap
        might be currently being loaded.
     *)
    ( try
	let bm = Hashtbl.find blockmaps id in
	bm # inactivate();
	Hashtbl.remove blockmaps id
      with
	| Not_found -> ()
    );
    ( try
	let bm_cache = Hashtbl.find blockmap_caches id in
	bm_cache # abort();
	Hashtbl.remove blockmap_caches id
      with
	| Not_found -> ()
    )

  val inode_locks = Hashtbl.create 79
  val owner_inodes = Hashtbl.create 73

  method lock_inode id owner =
    let own_inode() =
      let l = 
	try Hashtbl.find owner_inodes owner
	with Not_found -> [] in
      Hashtbl.replace owner_inodes owner (id :: l) in
    let free_inode() =
      try
	let l = 
	  Hashtbl.find owner_inodes owner in
	let l' = List.filter (fun i -> id <> i) l in
	if l' = [] then
	  Hashtbl.remove owner_inodes owner
	else
	  Hashtbl.replace owner_inodes owner l'
      with Not_found -> () in
    try
      let o = Hashtbl.find inode_locks id in
      if o = owner then Some (fun () -> ()) else None
    with
      | Not_found ->
	  Hashtbl.replace inode_locks id owner;
	  own_inode();
	  Some (fun () -> Hashtbl.remove inode_locks id; free_inode())

  method unlock_inodes owner =
    let inodes =
      try Hashtbl.find owner_inodes owner
      with Not_found -> [] in
    Hashtbl.remove owner_inodes owner;
    List.iter
      (fun id ->
	 try
	   let o = Hashtbl.find inode_locks id in
	   assert(o=owner);
	   Hashtbl.remove inode_locks id
	 with
	   | Not_found -> assert false
      )
      inodes


  val filename_locks = Hashtbl.create 51
  val owner_filenames = Hashtbl.create 53

  method lock_filename fn typ (owner : < > ) =
    let own_filename() =
      let l = 
	try Hashtbl.find owner_filenames owner
	with Not_found -> [] in
      if not (List.mem fn l) then
	Hashtbl.replace owner_filenames owner (fn :: l) in
    let free_filename() =
      try
	let l = 
	  Hashtbl.find owner_filenames owner in
	let l' = List.filter (fun n -> fn <> n) l in
	if l' = [] then
	  Hashtbl.remove owner_filenames owner
	else
	  Hashtbl.replace owner_filenames owner l'
      with Not_found -> () in
    let restore_lock old_lock_opt =
      match old_lock_opt with
	| None ->
	    free_filename();
	    Hashtbl.remove filename_locks fn
	| Some old ->
	    Hashtbl.replace filename_locks fn old in
    let unlock old_lock_opt () =
      (** First remove [owner] from the list of lock owners. If this results
          in an empty list, we can go back to the previous state
          ([old_lock_opt]).

          [unlock] is only designed to undo the last locking request, for the
          case that the operation fails.
       *)
      try
	match Hashtbl.find filename_locks fn with
	  | `Link o ->
	      assert(o = owner);
	      restore_lock old_lock_opt
	  | `Unlink(l,is_excl) ->
	      assert(List.mem owner l);
	      let l' = List.filter (fun o -> o <> owner) l in
	      if l' = [] then
		restore_lock old_lock_opt
	      else (
		free_filename();
		Hashtbl.replace filename_locks fn (`Unlink(l',is_excl))
	      )
	  | `Exists l ->
	      assert(List.mem owner l);
	      let l' = List.filter (fun o -> o <> owner) l in
	      if l' = [] then
		restore_lock old_lock_opt
	      else (
		free_filename();
		Hashtbl.replace filename_locks fn (`Exists l')
	      )
      with
	| Not_found ->
	    assert false in
    let set_lock old_lock_opt new_lock =
      Hashtbl.replace filename_locks fn new_lock;
      own_filename();
      Some (unlock old_lock_opt)
    in
    try
      let old_lock = Hashtbl.find filename_locks fn in
      match old_lock with
	| `Link o ->
	    if o <> owner then failwith "lock_filename: conflict";
	    (** If we own [`Link] we are allowed to do everything with the
                file. So we accept [`Exists] and [`Unlink] but do not 
                downgrade (which would allow competing [link] operations
                with uncertain semantics).
	     *)
	    Some (fun () -> ())
	| `Unlink([o], is_excl) when typ = `Link ->
	    if o <> owner then failwith "lock_filename: conflict";
	    (** change unlink->link *)
	    set_lock (Some old_lock) (`Link owner)
	| `Unlink(l, is_excl) ->
	    (** A change from `Unlink to `Exists is not reasonable. If the
                file was deleted, we can only create it by a link operation.
	     *)
	    if typ <> `Unlink || is_excl then
	      failwith "lock_filename: conflict";
	    let l' =
	      if not(List.mem owner l) then owner::l else l in
	    set_lock None (`Unlink(l',is_excl))
	| `Exists [o] when typ = `Unlink ->
	    (** change exists->unlink (normal when we delete files).
		We have to set the exclusive flag to true. This prevents
		that further [`Unlink] requests can be satisfied. This
		is required because there might otherwise no way back
		to the previous locking situation if the operation fails
		that needs the lock
	     *)
	    if o <> owner then failwith "lock_filename: conflict";
	    set_lock (Some old_lock) (`Unlink([owner],true))
	| `Exists l ->
	    (** A change from `Exists to `Link is nonsense. The latter
                guarentees exclusive creation of files.
	     *)
	    if typ <> `Exists then failwith "lock_filename: conflict";
	    let l' =
	      if not(List.mem owner l) then owner::l else l in
	    set_lock None (`Exists l')
    with
      | Not_found ->
	  (** If not found we always grant the lock. *)
	  let new_lock =
	    match typ with
	      | `Link -> `Link owner
	      | `Unlink -> `Unlink([owner],false)
	      | `Exists -> `Exists [owner] in
	  set_lock None new_lock
      | Failure _ ->
	  None

  method unlock_filenames owner =
    let locks =
      try Hashtbl.find owner_filenames owner
      with Not_found -> [] in
    Hashtbl.remove owner_filenames owner;
    List.iter
      (fun fn ->
	 try
	   match Hashtbl.find filename_locks fn with
	     | `Link o ->
		 assert(o = owner);
		 Hashtbl.remove filename_locks fn
	     | `Unlink(l,is_excl) ->
		 assert(List.mem owner l);
		 let l' = 
		   List.filter (fun o -> o <> owner) l in
		 if l' = [] then
		   Hashtbl.remove filename_locks fn
		 else
		   Hashtbl.replace filename_locks fn (`Unlink(l',is_excl))
	     | `Exists l ->
		 assert(List.mem owner l);
		 let l' = 
		   List.filter (fun o -> o <> owner) l in
		 if l' = [] then
		   Hashtbl.remove filename_locks fn
		 else
		   Hashtbl.replace filename_locks fn (`Exists l')
	 with
	   | Not_found ->
	       assert false
      )
      locks
end


(**********************************************************************)
(* block allocation w/ replication                                    *)
(**********************************************************************)

(* TODO: better allocation algorithms *)

(** Trivial algorithm: Do every block separately. TODO: Better
    implementation that allocates blocks in bigger units, and that
    tries to balance them
 *)

exception Skip_ds

let allocate_trivial_strategy blockmaps td ownobj repl index number =
  let blocks = ref [] in
  for k = 0 to number-1 do
    Netlog.logf `Debug
      "allocate_trivial_strategy: allocating block %d/%d" k number;
    let index_k =
      Int64.add index (Int64.of_int k) in
    (** So simply iterate until we have our stores... *)
    let d = ref 0 in
    ( try
	List.iter
	  (fun bm ->
	     let idstr = bm#identity in
	     Netlog.logf `Debug 
	       "allocate_trivial_strategy: looking at store %s" idstr;
	     try
	       (** Check whether bm is ok. *)
	       if not bm#has_free_blocks then (
		 Netlog.logf `Debug 
		   "allocate_trivial_strategy: store %s is full" idstr;
		 raise Skip_ds;
	       );
	       if not (Nn_datastore_news.is_alive idstr) then (
		 Netlog.logf `Debug 
		   "allocate_trivial_strategy: store %s is not alive" idstr;
		 raise Skip_ds
	       );
	       let ds = 
		 try
		   Nn_datastore_news.get bm#identity
		 with Not_found -> (
		 Netlog.logf `Debug 
		   "allocate_trivial_strategy: store %s is not monitored" idstr;
		   raise Skip_ds
		 ) in
	       let new_blocks = bm # reserve 1 ownobj in
	       let new_block = List.hd new_blocks in
	       let node =
		 match ds.Nn_datastores.node  with
		   | Some addr ->
		       let (h,p) = Plasma_util.host_port_of_sockaddr addr in
		       sprintf "%s:%d" h p
		   | None ->
		       Netlog.logf `Debug 
			 "allocate_trivial_strategy: \
                          store %s has weird address" idstr ;
		       raise Skip_ds in
	       let vfy =
		 Plasma_util.compute_verifier 
		   td.td_safetrans_id 
		   td.td_safetrans_secret
		   new_block in
	       let bi =
		 { Pfs_rpcapi_aux.index = index_k;
		   node = node;
		   node_alive = true;
		   identity = bm#identity;
		   block = new_block;
		   checksum = None;
		   inode_seqno = 0L;  (* not known here *)
		   safetrans_id = td.td_safetrans_id;
		   safetrans_tmo = td.td_safetrans_tmo;
		   safetrans_vfy = vfy
		 } in
	       blocks := bi :: !blocks;
	       incr d;
	       if !d >= repl then raise Exit
	     with
	       | Skip_ds -> ()
	       | Nn_blockmap.Inactive -> 
		   (** destroyed blockmap *)
		   Netlog.logf `Debug 
		     "allocate_trivial_strategy: \
                      store %s is inactive" idstr ;
	  )
	  blockmaps
      with
	| Exit -> ()
    );
    if !d < repl then (
      Netlog.logf `Debug
	"allocate_trivial_strategy: \
         found blocks in %d stores (%d expected) - ENOSPC"
	!d repl;
      raise(Transaction_error `enospc)
    )
  done;
  List.rev !blocks
    

let allocate_enblock blockmaps td ownobj repl index number pref =
  (** Sort the datastores by free blocks. Try to get [number] blocks from
      a single datastore (with most free blocks). If this works, go on,
      and get further [number] blocks for the replicas from other datastores.

      If there is no datastore with enough blocks, we have to split the
      allocation request: We get as many as we can, and continue our
      algorithm with this allocation as above. For the missing blocks
      we also recurse.
   *)

  (** First sort datastores by free blocks: *)
  let good_blockmaps =
    List.flatten
      (List.map
	 (fun bm ->
	    try
	      let idstr = bm#identity in
	      if not bm#has_free_blocks then (
		Netlog.logf `Debug 
		  "allocate_enblock: store %s is full" idstr;
		raise Skip_ds;
	      );
	      if not (Nn_datastore_news.is_alive idstr) then (
		Netlog.logf `Debug 
		  "allocate_enblock: store %s is not alive" idstr;
		raise Skip_ds
	      );
	      let ds = 
		try
		  Nn_datastore_news.get bm#identity
		with Not_found -> (
		  Netlog.logf `Debug 
		    "allocate_enblock: store %s is not monitored" idstr;
		  raise Skip_ds
		) in
	      let (h,p) =
		match ds.Nn_datastores.node  with
		  | Some addr ->
		      Plasma_util.host_port_of_sockaddr addr
		  | None ->
		      Netlog.logf `Debug 
			"allocate_enblock: \
                          store %s has weird address" idstr ;
		      raise Skip_ds in
	      let node =
		sprintf "%s:%d" h p in
	      let is_preferred =
		List.mem idstr pref in
	      [ bm, node, is_preferred ]
	    with
	      | Skip_ds -> []
	 )
	 blockmaps) in

  let sorted_blockmaps =
    List.sort
      (fun (bm1,_,pr1) (bm2,_,pr2) -> 
	 (** Sort preferred datastores to the beginning. Second criterion
	     is the percentage of free blocks (biggest first)
	  *)
	 let q =
	   match pr1 with
	     | true -> if pr2 then 0 else (-1)
	     | false -> if pr2 then 1 else 0 in
	 match q with
	   | 0 ->
	       compare
		 (Int64.to_float bm2#n_free /. Int64.to_float bm2#size)
		 (Int64.to_float bm1#n_free /. Int64.to_float bm1#size)
	   | _ -> 
	       q
      )
      good_blockmaps in

  let create_bi bm node index blocknr =
    let vfy =
      Plasma_util.compute_verifier 
	td.td_safetrans_id 
	td.td_safetrans_secret
	blocknr in
    { Pfs_rpcapi_aux.index = index;
      node = node;
      node_alive = true;
      identity = bm#identity;
      block = blocknr;
      checksum = None;
      inode_seqno = 0L;  (* not known here *)
      safetrans_id = td.td_safetrans_id;
      safetrans_tmo = td.td_safetrans_tmo;
      safetrans_vfy = vfy
    } in

  let rec create_bi_list bm node index blocknr_list =
    match blocknr_list with
      | blocknr :: blocknr_list' ->
	  create_bi bm node index blocknr :: 
	    create_bi_list bm node (Int64.succ index) blocknr_list'
      | [] ->
	  [] in

  let rec allocate_range used_set index number repl =
    (** Still have to get [repl] replicas for the range 
	[index] to [index+number-1]. The datastores in [used_set]
	have already replicas, and cannot be used.
     *)
    if repl = 0 then
      []
    else
      let avail_blockmaps =
	List.filter
	  (fun (bm,_,_) ->
	     bm#has_free_blocks && not (StrSet.mem bm#identity used_set)
	  )
	  sorted_blockmaps in
      match avail_blockmaps with
	| (bm,node,_) :: _ ->
	    let n = 
	      Int64.to_int (min (Int64.of_int number) bm#n_free) in
	    let blocks = bm#reserve n ownobj in
	    let bi_list = create_bi_list bm node index blocks in
	    let used_set' = StrSet.add bm#identity used_set in
	    if n = number then
	      bi_list @ 
		allocate_range used_set' index number (repl-1)
	    else (
	      let index1 =
		Int64.add index (Int64.of_int n) in
	      let n1 =
		number - n in
	      bi_list @
		allocate_range used_set' index n (repl-1) @
		allocate_range used_set  index1 n1 repl
	    )
	| [] ->
	    raise(Transaction_error `enospc)
  in

  allocate_range StrSet.empty index number repl


let alloc_limit = 64 * 1024 * 1024

let rec allocate_balanced_strategy_1 blockmaps td ownobj repl index number 
                                     blocksize pref =
  (** Use allocate_enblock in units of:
      - no more blocks than bmaprowsize
      - as many blocks as fit in 64M 
      - at least one block
   *)
  let bmaprowsize = Nn_config.bmaprowsize in
  let rows_by_limit = max 1 (alloc_limit / blocksize) in
  let allocsize = min bmaprowsize rows_by_limit in
  let allocsizeL = Int64.of_int allocsize in
  if number <=  allocsize then
    allocate_enblock blockmaps td ownobj repl index number pref
  else
    allocate_enblock blockmaps td ownobj repl index allocsize pref @
      (allocate_balanced_strategy_1
         blockmaps td ownobj repl
         (Int64.add index allocsizeL) (number-allocsize) blocksize pref)

let allocate_balanced_strategy blockmaps td ownobj repl index number 
                               blocksize pref =
  let l = 
    allocate_balanced_strategy_1
      blockmaps td ownobj repl index number blocksize pref in
  assert(List.length l = repl * number);
  l


(**********************************************************************)
(* Datastore view                                                     *)
(**********************************************************************)


(** The datastore view always loads the datastore table anew from db
    when it is established. This view is only needed for doing
    certain modifications of the datastore table. No need to implement
    a cache.

    However, when we disable a datastore (or even delete it), we have
    to propagate this to the various caches.
 *)

class ds_view (shared_state:shared_state_t) 
           : ds_view_t =
  let () = (
    if not (!(shared_state#master_enabled)) then
      failwith "Nn_state: The master db is out of sync. No access permitted"
  ) in
  let esys = Nn_db.ro_esys() in
  let datastore_cache =  (** Cached [datastore] table *)
    Uq_engines.cache
      (fun esys ->
	 Nn_db.datastore_list_ro_e()
	 ++ (fun l ->
	       (* Update [shared_state] on the fly *)
	       shared_state#datastore_use_id 0;
	       List.iter
		 (fun ds -> 
		    shared_state#datastore_use_id ds.Nn_datastores.id)
		 l;
	       eps_e (`Done l) esys
	    )
      )
      esys in
  let journal = Queue.create() in     (** Queue of updates *)
  let to_forget = Queue.create() in   (** Identities to fortget after commit *)
object(self)

  method datastore_list_e() =
    datastore_cache # get_engine()

  method datastore_new_id_e() =
    (** Check whether [shared_state] already knows the maximum ID. If so, just
        increase by 1. Otherwise, get from [datastore_cache] once which will
        update the maximum ID
     *)
    match shared_state # datastore_max_id with
      | None ->
	  let e = datastore_cache # get_engine() in
	  new Uq_engines.map_engine
	    ~map_done:(fun _ -> 
			 (** Now we can safely assume that max ID is known *)
			 match shared_state # datastore_max_id with
			   | None ->
			       assert false
			   | Some m ->
			       let id = m+1 in
			       shared_state # datastore_use_id id;
			       `Done id
		      )
	    e
      | Some m ->
	  (** max is m *)
	  let id = m+1 in
	  shared_state # datastore_use_id id;
	  new Uq_engines.epsilon_engine (`Done id) esys

  method datastore_upd_e ~id ~identity ~size ~enabled =
    (** First ensure we know the newest value. Then update the value,
        and record the update in our journal.
     *)
    new Uq_engines.map_engine
      ~map_done:(fun l ->
		   try
		     let have_id =
		       List.exists
			 (fun ds -> ds.Nn_datastores.id = id) l in
		     if have_id then (
		       (* it's an update *)
		       let ds =
			 List.find
			   (fun ds -> ds.Nn_datastores.id = id) l in
		       if ds.Nn_datastores.identity <> identity then
			 failwith "Nn_state.datastore_upd_e: cannot change \
                                   identity";
		       if ds.Nn_datastores.size > size then
			 failwith "Nn_state.datastore_upd_e: cannot shrink \
                                   datastore size";
		       let ds' =
			 { ds with
			     Nn_datastores.size = size;
			     enabled = enabled
			 } in
		       let l' =
			 List.map
			   (fun ds ->
			      if ds.Nn_datastores.id = id then ds' else ds)
			   l in
		       datastore_cache # put l';
		       if not enabled && ds.Nn_datastores.enabled then
			 Queue.push identity to_forget
		     )
		     else (
		       (* it's an addition *)
		       let have_identity =
			 List.exists
			   (fun ds -> ds.Nn_datastores.identity = identity) 
			   l in
		       if have_identity then
			 failwith "Nn_state.datastore_upd_e: not unique";
		       let ds =
			 { Nn_datastores.id = id;
			   identity = identity;
			   size = size;
			   enabled = enabled;
			   node = None
			 } in
		       let l' = ds :: l in
		       datastore_cache # put l'
		     );
		     (* Finally remember the update so we can push it to
                        the database and the slaves later
		      *)
		     let je =
		       `Datastore_upd(id,identity,size,enabled) in
		     Queue.push je journal;
		     `Done()
		   with
		     | error ->
			 `Error error
		)
      (datastore_cache # get_engine())

  method datastore_del_e ~id  =
    (** First ensure we know the newest value. Then record the delete in
        our journal.
     *)
    new Uq_engines.map_engine
      ~map_done:(fun l ->
		   let have_id =
		     List.exists
		       (fun ds -> ds.Nn_datastores.id = id) l in
		   if have_id then (
		     let ds =
		       List.find
			 (fun ds -> ds.Nn_datastores.id = id) l in
		     Queue.push (`Datastore_del id) journal;
		     Queue.push ds.Nn_datastores.identity to_forget;
		     `Done ()
		   )
		   else
		     `Error (Transaction_error `enonode)
		)
      (datastore_cache # get_engine())

  method pre_commit =
    journal

  method post_commit ok =
    if ok then (
      (** Call [forget_identity] for all disabled or deleted datastores.
          This is important to ensure that at enable time the caches
          referring to this datastore are loaded from db
       *)
      Queue.iter
	(fun identity ->
	   shared_state # forget_identity identity
	)
	to_forget;
    )

  method rollback() =
    ()
end


(**********************************************************************)
(* Blockmap view                                                      *)
(**********************************************************************)

(** The blockmap view maintains a shared cache (in [shared_state_t]).
 *)

class bm_view (shared_state:shared_state_t)
           : bm_view_t =
  let () = (
    if not (!(shared_state#master_enabled)) then
      failwith "Nn_state: The master db is out of sync. No access permitted"
  ) in
  let esys = Nn_db.ro_esys() in
  let journal = Queue.create() in
object(self)

  method blockmap_get_e ~identity =
    shared_state # blockmap_get_e ~identity

  method private blockmap_mget_e identities =
    (** Start an engine getting the blockmaps for [identities]. If
        this list is empty, get the blockmaps for all enabled datastores.
	When the engine is done, the blockmaps can be retrieved directly
        via [shared_state#blockmaps].

        This method fails if there is an invalid identity, or an
        identity is destroyed while loading. Failure results in an
        engine transitioning to an [`Error] or [`Aborted] state.
     *)
    let identities' =
      if identities = [] then
	let ds_list =
	  Nn_datastore_news.list ~wait:true () in
	List.map
	  (fun ds -> ds.Nn_datastores.identity)
	  (List.filter
	     (fun ds -> ds.Nn_datastores.enabled)
	      ds_list)
      else
	identities in
    Netlog.logf `Debug
      "blockmap_mget_e: loading datastores %s"
      (String.concat "," identities);
    Plasma_util.msync_engine 
      (List.map
	 (fun identity -> 
	    new Uq_engines.map_engine
	      ~map_done:(fun _ -> `Done())
	      (self # blockmap_get_e ~identity))
	 identities'
      )
      esys

  method blockmap_alloc_e ~rpolicy ~index ~number ~pref td =
    let `Balanced repl = rpolicy in
    new Uq_engines.map_engine
      ~map_done:(fun () ->
		   try
		     let blockmaps = shared_state#blockmaps in
		     let td_blockmaps =
		       List.filter
			 (fun bm -> List.mem bm#id td.td_idlist)
			 blockmaps in
		     let blocksize = shared_state#node_config#dn_blocksize in
		     let owner = (self :> < >) in
		     let blocks =
		       (* allocate_trivial_strategy  *)
		       allocate_balanced_strategy
			 td_blockmaps td owner repl index number blocksize 
			 pref in
		     `Done (Array.of_list blocks)
		   with
		     | error ->
			 `Error error
		)
      (self#blockmap_mget_e [])

  method blockmap_dealloc_e blocks =
    (** First get all blockmaps we need: *)
    let ids = ref StrSet.empty in
    List.iter
      (fun (id,_) -> ids := StrSet.add id !ids)
      blocks;
    let ids_l = StrSet.fold (fun id acc -> id::acc) !ids [] in
    let e =
      self#blockmap_mget_e
	ids_l in
    Uq_engines.when_state
      ~is_done:(fun () ->
		  List.iter
		    (fun (id,block) ->
		       let bm = shared_state#get_blockmap id in
		       bm#free [block] (self :> < >)
		    )
		    blocks
	       )
      e;
    e

  method blockmap_load_e () =
    (** First get all known blockmap identities: *)
    let ids = ref StrSet.empty in
    List.iter
      (fun ds -> ids := StrSet.add ds.Nn_datastores.identity !ids)
      (Nn_datastore_news.list ~wait:true ());
    let ids_l = StrSet.fold (fun id acc -> id::acc) !ids [] in
    self#blockmap_mget_e
      ids_l

  method blockmap_keep blocks =
    List.iter
      (fun (id,block) ->
	 try
	   let bm = shared_state#get_blockmap id in
	   bm#keep [block] (self :> < >)
	 with
	   | Not_found ->
	       failwith "blockmap_keep: blockmap not loaded"
      )
      blocks


  method pre_commit =
    (** Check all blockmaps, and extract db changes *)
    List.iter
      (fun bm ->
	 let id = bm # id in
	 let l = bm#get_changes (self :> < >) in
	 List.iter
	   (fun (index,data) ->
	      Queue.push (`Blockalloc_upd(id,index,data)) journal
	   )
	   l
      )
      shared_state#blockmaps;
    journal


  method post_commit ok =
    if ok then (
      (** Make changes permanent:
	  (Note that this [commit] is the one that must not overlap
	  with [inode_get_blocks_e].)
       *)
      List.iter
	(fun bm -> bm#commit (self :> < >))
	shared_state#blockmaps;
    )
    else (
      (** Unreserve all blocks: *)
      List.iter
	(fun bm -> bm#rollback (self :> < >))
	shared_state#blockmaps;
    )


  method rollback () =
    self#post_commit false
end


(**********************************************************************)
(* Inode view                                                         *)
(**********************************************************************)

type inode_detail =
    { i_info : Pfs_rpcapi_aux.inodeinfo;
      i_blocks : (int64, Pfs_rpcapi_aux.blockinfo list option) Hashtbl.t
    }

type inode =
    { i_id : int64;
      i_detail : inode_detail option; (* None: deleted *)
    }


class inode_view (shared_state:shared_state_t) 
           : inode_view_t =
  let () = (
    if not (!(shared_state#master_enabled)) then
      failwith "Nn_state: The master db is out of sync. No access permitted"
  ) in
  let esys = Nn_db.ro_esys() in
  let journal = Queue.create() in
object(self)
  val inodes = Hashtbl.create 7
    (** Cached updates of this transaction *)

  val inode_links = Hashtbl.create 7
    (** This number [n] means that we now about at least [n] links for the
	inode. If [n=0] we have to check at commit time whether the inode
	needs to be deleted. Enter here only something if we also have a
	lock for the inode.
     *)

  method private modify_link_number inode f =
    let n = try Hashtbl.find inode_links inode with Not_found -> 0 in
    let n' = f n in
    Hashtbl.replace inode_links inode n'
    

  method inode_get_e ~id =
    try
      let i = Hashtbl.find inodes id in
      let r =
	match i.i_detail with
	  | None -> None
	  | Some d -> Some d.i_info in
      new Uq_engines.epsilon_engine (`Done r) esys
    with
      | Not_found ->
	  Nn_db.inode_get_ro_e ~id

  method inode_get_blocks_e ~id ~blkidx ~len bmv =
    (** First try to get blocks from the cache [inodes]. If blocks are
        missing, do the db query, and add missing blocks from there

	The difficulty here is to immediately set the keep flag in the
	block map - before a competing transaction can commit. Because of
	this we first load all blockmaps
     *)
    let keep bis =
      bmv # blockmap_keep
	(List.map
	   (fun bi ->
	      (bi.Pfs_rpcapi_aux.identity, bi.Pfs_rpcapi_aux.block)
	   )
	   bis
	)
    in
    bmv # blockmap_load_e ()
    ++ (fun () ->
	  let known = ref [] in
	  let missing = Hashtbl.create 7 in
	  let missing_flag = ref false in
	  try
	    let i = Hashtbl.find inodes id in
	    missing_flag := true;
	    match i.i_detail with
	      | None ->
		  (** inode was deleted *)
		  new Uq_engines.epsilon_engine (`Done []) esys
	      | Some d ->
		  let b = ref blkidx in
		  let b_end = Int64.add blkidx len in
		  while !b < b_end do
		    ( try
			let bi_opt = Hashtbl.find d.i_blocks !b in
			match bi_opt with
			  | None ->
			      (** block was deleted *)
			      ()
			  | Some bis ->
			      keep bis;
			      known := (List.rev bis) @ !known
		      with
			| Not_found ->
			    (** block was not updated *)
			    Hashtbl.replace missing !b ()
		    );
		    b := Int64.succ !b
		  done;
		  (** if we need a db lookup, raise Not_found *)
		  if Hashtbl.length missing > 0 then
		    raise Not_found
		  else (
		    eps_e (`Done (List.rev !known)) esys
		  )
	  with
	    | Not_found ->
		(** Do the db lookup, optionally followed by postprocessing.
		    The db lookup and the following [keep] call must not
		    overlap with a blockmap commit. Because of this it is
		    run from a prioritizer.
		 *)
		shared_state # blockmap_prioritizer # prioritized
		  (fun _ ->
		     Nn_db.inodeblocks_get_ro_e ~inode:id ~blkidx ~len 
		     ++ (fun bis ->
			   if !missing_flag then (
			     (** Only take those rows from the db where
				 [missing] says it
			      *)
			     let l = ref [] in
			     List.iter
			       (fun bi ->
				  let p_miss =
				    Hashtbl.mem
				      missing bi.Pfs_rpcapi_aux.index in
				  if p_miss then
				    l := bi :: !l
			       )
			       bis;
			     keep !l;
			     known := List.rev (!l @ !known);
			     eps_e (`Done()) esys
			   )
			   else (
			     (** no entry in the cache anyway, so just return *)
			     known := bis;
			     eps_e (`Done()) esys
			   )
			)
		  )
		  2
		++ (fun () ->
		      eps_e (`Done !known) esys
		   )
       )

  method inode_alloc_e ii =
    (** First ensure that we know the maximum ID. This access is serialized
        so it cannot overlap with the insertion of a new ID
     *)
    new Uq_engines.seq_engine
      (shared_state # inode_get_new_id_serialized
	 (fun esys ->
	    match shared_state#inode_max_id with
	      | None ->
		  new Uq_engines.map_engine
		    ~map_done:(fun m ->
				 let id = Int64.succ m in
				 shared_state # inode_use_id id;
				 `Done id
			      )
		    (Nn_db.inode_maxid_ro_e ())
	      | Some m ->
		  let id = Int64.succ m in
		  shared_state # inode_use_id id;
		  new Uq_engines.epsilon_engine (`Done id) esys
	 ))
      (fun id ->
	 let ii' = { ii with Pfs_rpcapi_aux.blocklimit = 0L } in
	 Queue.push (`Inode_ins(id,ii')) journal;
	 let d = { i_info = ii'; i_blocks = Hashtbl.create 7 } in
	 let i = { i_id = id; i_detail = Some d } in
	 Hashtbl.add inodes id i;
	 let unlock_opt =
	   shared_state # lock_inode id (self :> < > ) in
	 assert(unlock_opt <> None);   (* the inode is new! *)
	 self # modify_link_number id (fun _ -> 0);
	 new Uq_engines.epsilon_engine (`Done id) esys
      )

  method private update_e ~id ~blocklimit ~set_mtime ~set_ctime
                           ii_opt new_bis del_blkidx del_len =
    (** Combined update: if [ii_opt] is set, this is the new inodeinfo.
        The blocks in [new_bis] are added to the blocklist.
        The blocks [del_blkidx] to [del_blkidx+len-1] are deleted.

	The inodeinfo is now always updated - at least seqno is incremented.
	If [set_mtime], the mtime field is set to server time. 
	If [set_ctime], the ctime field is set to server time. 
	[blocklimit=`Ignore]: New values in [ii_opt] are ignored.
	[blocklimit=`Set n]: The blocklimit is set to [n].
	[blocklimit=`Update]: The blocklimit is updated as in [ii_opt]
     *)
    let unlock_opt =
      shared_state # lock_inode id (self :> < >) in
    match unlock_opt with
      | None ->
	  new Uq_engines.epsilon_engine
	    (`Error (Transaction_error `econflict)) esys
      | Some unlock ->
	  (** First get the inodeinfo *)
	  let e =
	    new Uq_engines.seq_engine
	      (self # inode_get_e ~id)
	      (function
		 | None ->
		     new Uq_engines.epsilon_engine
		       (`Error(Transaction_error `enoient)) esys
		 | Some old_ii ->
		     (** Get new i_detail with the ii_opt update (if any) *)
		     let new_ii0 =
		       match ii_opt with
			 | None -> old_ii
			 | Some ii -> ii  in
		     let now = Unix.gettimeofday() in
		     let new_ii =
		       { new_ii0 with 
			   Pfs_rpcapi_aux.blocklimit = (
			     match blocklimit with
			       | `Ignore ->
				   old_ii.Pfs_rpcapi_aux.blocklimit 
			       | `Set n ->
				   n
			       | `Update ->
				   new_ii0.Pfs_rpcapi_aux.blocklimit
			   );
			   mtime = (
			     if set_mtime then
			       encap_time now
			     else
			       new_ii0.Pfs_rpcapi_aux.mtime);
			   ctime = (
			     if set_ctime then
			       encap_time now
			     else
			       new_ii0.Pfs_rpcapi_aux.ctime);
			   seqno = Int64.succ old_ii.Pfs_rpcapi_aux.seqno
		       } in
		     let i_detail =
		       try
			 let i_old = Hashtbl.find inodes id in
			 match i_old.i_detail with
			   | None ->
			       { i_info = new_ii; i_blocks = Hashtbl.create 7 }
			   | Some d ->
			       { d with i_info = new_ii }
		       with
			 | Not_found -> 
			     { i_info = new_ii; i_blocks = Hashtbl.create 7 } in
		     (** Delete the blocks: *)
		     let b = ref del_blkidx in
		     let b_end = Int64.add del_blkidx del_len in
		     while !b < b_end do
		       ( try
			   let old_bis_opt =
			     Hashtbl.find i_detail.i_blocks !b in
			   if old_bis_opt <> None then
			     Hashtbl.replace i_detail.i_blocks !b None
			 with Not_found -> ()
		       );
		       b := Int64.succ !b;
		     done;
		     (** Now apply new_bis: *)
		     List.iter
		       (fun bi ->
			  let idx = bi.Pfs_rpcapi_aux.index in
			  try
			    let old_bis_opt =
			      Hashtbl.find i_detail.i_blocks idx in
			    match old_bis_opt with
			      | None ->
				  (** Block was deleted. Overwrite *)
				  raise Not_found
			      | Some old_bis ->
				  (** Add new block *)
				  Hashtbl.replace 
				    i_detail.i_blocks
				    idx
				    (Some (bi :: old_bis))
			  with
			    | Not_found ->
				Hashtbl.replace 
				  i_detail.i_blocks
				  idx
				  (Some [ bi ])
		       )
		       new_bis;
		     (** Update the journal: *)
		     Queue.push (`Inode_upd(id,new_ii)) journal;
		     if del_len > 0L then
		       Queue.push
			 (`Inodeblocks_del(id,del_blkidx,del_len))
			 journal;
		     Queue.push (`Inodeblocks_ins(id,new_bis)) journal;
		     (** Update the cache [inodes] *)
		     let i = { i_id = id; i_detail = Some i_detail } in
		     Hashtbl.replace inodes id i;
		     new Uq_engines.epsilon_engine (`Done()) esys
	      ) in
	  Uq_engines.when_state
	    ~is_error:(fun _ -> unlock())  ~is_aborted:unlock
	    e;
	  e

  method inode_upd_e ~id ~keep_blocklimit ii =
    let blocklimit =
      if keep_blocklimit then `Ignore else `Update in
    self # update_e
      ~id ~blocklimit ~set_mtime:false ~set_ctime:false
      (Some ii) [] 0L 0L

  method inode_upd_blocks_e ~id ~del ~add ~set_mtime ~set_blocklimit =
    let (del_blkidx, del_len) = del in
    let blocklimit =
      if set_blocklimit >= 0L then
	`Set set_blocklimit
      else
	`Ignore in
    self # update_e
      ~id ~blocklimit ~set_mtime ~set_ctime:false
      None add del_blkidx del_len

  method inode_upd_time ~id ~mtime ~ctime =
    Queue.push
      (`Inode_upd_time(id, mtime, ctime))
      journal
    

  method inode_dealloc_e ~id =
    (** We only have to check whether the row really exists *)
    let unlock_opt = shared_state # lock_inode id (self :> < >) in
    match unlock_opt with
      | None ->
	  new Uq_engines.epsilon_engine
	    (`Error (Transaction_error `econflict)) esys
      | Some unlock ->
	  let e =
	    self # inode_get_e ~id
	    ++ (function
		  | None ->
		      transerr_e `enoient esys
		  | Some _ ->
		      eps_e (`Done ()) esys
	       )
	    ++ (fun () ->
		  (** Check that there are no filenames: *)
		  self # filename_rev_get_e ~inode:id
	       )
	    ++ (fun names ->
		  if names <> [] then
		    transerr_e `efhier esys
		  else (
		    Queue.push (`Inode_del id) journal;
		    let i = { i_id = id; i_detail = None } in
		    Hashtbl.replace inodes id i;
		    eps_e (`Done ()) esys
		  )
	       ) in
	  Uq_engines.when_state
	    ~is_error:(fun _ -> unlock())  ~is_aborted:unlock
	    e;
	  e

  (** filenames *)

  val mutable fnupd_by_name = Hashtbl.create 7
    (** Cached updates by name. This is [None] if the name was deleted,
        and [Some i] when it was added with inode [i]
     *)

  val mutable fnupd_by_inode = Hashtbl.create 7
    (** Cached updates by inode. This is a pair of two lists [(add,del)]
        where [add] is a list of names to which the inode was added and
        [del] as list of names from which the inode was deleted. Both lists
        do not intersect

	TODO: add, del could be sets
     *)

  method filename_get_e ~name =
    try
      let inode_opt = Hashtbl.find fnupd_by_name name in
      eps_e (`Done inode_opt) esys
    with
      | Not_found ->
	  Nn_db.names_get_ro_e ~name

  method filename_rev_get_e ~inode =
    let add_names,del_names = 
      try Hashtbl.find fnupd_by_inode inode
      with Not_found -> [], [] in
    Nn_db.names_rev_get_ro_e ~inode
    ++ (fun names ->
	  let names' =
	    add_names @ 
	      (List.filter (fun n -> not(List.mem n del_names)) names) in
	  (** Not all unlinks are represented in [fnupd_by_inode]. Also check
              with [fnupd_by_name]:
	   *)
	  let names'' =
	    List.filter
	      (fun n -> 
		 try Hashtbl.find fnupd_by_name n <> None 
		 with Not_found -> true
	      )
	      names' in
	  eps_e (`Done names'') esys
       )

  method filename_count_e ~inode =
    (* TODO: we can implement this better! *)
    self # filename_rev_get_e ~inode
    ++ (fun names ->
	  eps_e (`Done(List.length names)) esys
       )

  method filename_list_e ~name =
    let name0 = if name = "/" then "" else name in
    let name1 = name0 ^ "/" in
    let l1 = String.length name1 in
    (** First check whether [name] is deleted. *)
    let dir_ok =
      try Hashtbl.find fnupd_by_name name <> None
      with Not_found -> true in
    if dir_ok then (
      Nn_db.names_list_ro_e ~name
      ++ (fun matches ->
	    Netlog.logf `Debug "filename_list(%s)=%s" 
	      name (String.concat "," (List.map fst matches));
	    (** First check whether matches have been deleted *)
	    let matches' =
	      List.filter
		(fun (m,_) ->
		   try Hashtbl.find fnupd_by_name (name1 ^ m) <> None
		   with Not_found -> true
		)
		matches in
	    (** Find additional names. This is a bit ugly. (TODO) *)
	    let matches'' =
	      Hashtbl.fold
		(fun n d acc ->
		   match d with
		     | Some inode ->
			 if String.length n >= l1 && String.sub n 0 l1 = name1
			 then
			   let p = String.sub n l1 (String.length n - l1) in
			   if p <> "" && not (String.contains p '/') && 
			     not(List.exists (fun (q,_) -> q=p) acc)
			   then (p,inode) :: acc
			   else acc
			 else acc
		     | None ->
			 acc
		)
		fnupd_by_name
		matches' in
	    eps_e (`Done matches'') esys
	 )
    )
    else
      eps_e (`Done []) esys


  method filename_link_e ~name ~parent ~inode =
    let acquire_locks() =
      let owner = (self :> < > ) in
      (** We have to get three locks:
	  - [`Link] for filename [name]
	  - [`Exists] for filename [parent]
	  - a lock for the inode
       *)
      match shared_state # lock_filename name `Link owner with
	| None ->
	    None
	| Some unlock_name ->
	    ( match shared_state # lock_filename parent `Exists owner with
		| None ->
		    unlock_name(); None
		| Some unlock_parent ->
		    ( match shared_state # lock_inode inode owner with
			| None ->
			    unlock_name(); unlock_parent(); None
			| Some unlock_inode ->
			    let unlock() =
			      unlock_name(); unlock_parent(); unlock_inode() in
			    Some unlock
		    )
	    )
    in
    let push parent_inode =
      (** Do the link operation: *)
      Hashtbl.replace fnupd_by_name name (Some inode);
      let add,del = 
	try Hashtbl.find fnupd_by_inode inode
	with Not_found -> [], [] in
      let add' = name :: add in
      let del' = List.filter (fun n -> n <> name) del in
      Hashtbl.replace fnupd_by_inode inode (add', del');
      Queue.push
	(`Names_ins (name,inode))
	journal;
      self # modify_link_number inode (fun n -> n+1);
      eps_e (`Done parent_inode) esys in
    match acquire_locks() with
      | None ->
	  eps_e (`Error(Transaction_error `econflict)) esys
      | Some unlock ->
	  let e =
	    (** Check: The name must be new *)
	    self # filename_get_e ~name 
	    ++ (fun old_inode_opt ->
		  if old_inode_opt <> None then
		    transerr_e `eexist esys
		  else (
		    (** Get the inodeinfo of the file to link: *)
		    self # inode_get_e ~id:inode
		  )
	       )
	    ++ (fun ii_opt ->
		  (** If the inode is a directory, we have to check
		      that there is no other name pointing to it
		   *)
		  match ii_opt with
		    | None ->
			transerr_e `enoient esys
		    | Some ii ->
			match ii.Pfs_rpcapi_aux.filetype with
			  | `ftype_directory ->
			      (** Get other names of the directory *)
			      self # filename_rev_get_e ~inode 
			      ++ (fun names ->
				    if names <> [] then
				      transerr_e `efhier esys
				    else
				      (** All ok *)
				      eps_e (`Done ()) esys
				 )
			  | _ -> 
			      (** no directory: any number of links allowed *)
			      eps_e (`Done ()) esys
	       )
	    ++ (fun () ->
		  (** Look up the designated parent directory *)
		  self # filename_get_e ~name:parent
	       )
	    ++ (fun parent_inode_opt ->
		  match parent_inode_opt with
		    | None ->
			Netlog.logf `Debug "link: parent not found: %s"
			  parent;
			transerr_e `efhier esys
		    | Some parent_inode ->
			(** Get the parent inodeinfo: *)
			self # inode_get_e ~id:parent_inode
			++ (fun parent_ii_opt -> 
			      eps_e (`Done(parent_inode,parent_ii_opt)) esys
			   )
	       )
	    ++ (fun (parent_inode,parent_ii_opt) ->
		  match parent_ii_opt with
		    | None ->
			(** return here enoent not enoient because the user
			   referenced the parent dir by name
			 *)
			transerr_e `enoent esys
		    | Some parent_ii ->
			(** Check whether the parent is a directory *)
			let ft =
			  parent_ii.Pfs_rpcapi_aux.filetype in
			if ft <> `ftype_directory then
			  transerr_e `efhier esys
			else
			  push parent_inode
	       ) in
	  Uq_engines.when_state
	    ~is_error:(fun _ -> unlock()) ~is_aborted:unlock e;
	  e
	    
  method private push_unlink_e ~name ~inode n_links =
    (** We do not check here whether [name] really exists in the db.
        Just delete if it exists.
     *)
    let known_inode_opt = 
      try Hashtbl.find fnupd_by_name name
      with Not_found -> None in
    Hashtbl.replace fnupd_by_name name None;
    ( match known_inode_opt with
	| Some inode ->
	    let add,del = 
	      try Hashtbl.find fnupd_by_inode inode
	      with Not_found -> [], [] in
	    let add' = List.filter (fun n -> n <> name) add in
	    let del' = if List.mem name del then del else name :: del in
	    Hashtbl.replace fnupd_by_inode inode (add', del');
	| None ->
	    (** This is ok, but the cost not to update [fnupd_by_inode] is
                that there is another check in [filename_rev_get_e].
	     *)
	    ()
    );
    Queue.push
      (`Names_del name)
      journal;
    self # modify_link_number inode (fun n -> n_links);
    self # return_parent_inode_e ~name

  method private return_parent_inode_e ~name =
    try
      let last_slash = String.rindex name '/' in
      let parent_name = 
	if last_slash = 0 then "/" else String.sub name 0 last_slash in
      self # filename_get_e ~name:parent_name
      ++ (function
	    | None -> assert false
	    | Some inode -> eps_e (`Done inode) esys
	 )
    with
      | _ -> assert false

  method filename_unlink_e ~name =
    let owner = (self :> < > ) in
    let unlock_i = ref (fun () -> ()) in
    match shared_state#lock_filename name `Unlink owner with
      | None ->
	  transerr_e `econflict esys
      | Some unlock ->
	  (** First look the inode up (want to know the type). *)
	  let e =
	    self # filename_get_e ~name 
	    ++ (fun inode_opt ->
		  match inode_opt with
		    | None ->
			transerr_e `enoent esys
		    | Some inode ->
			(** Have to lock here the inode, unfortunately.
			    (TODO: get rid of this lock.) We do this here
			    so we can always delete the inode as post
			    operation if the number of links drops to 0
			 *)
			match shared_state # lock_inode inode owner with
			  | None ->
			      transerr_e `econflict esys
			  | Some unlock_inode ->
			      unlock_i := unlock_inode;
			      (** Get the inodeinfo *)
			      self # inode_get_e ~id:inode
			      ++ (fun ii_opt ->
				    eps_e (`Done(inode,ii_opt)) esys
				 )
	       )
	    ++ (fun (inode,ii_opt) ->
		  match ii_opt with
		    | None -> (* strange: filename without inode *)
			assert false
		    | Some ii ->
			(** Get the number of links: *)
			self # filename_rev_get_e ~inode
			++ (fun names ->
			      eps_e (`Done(inode,ii,names)) esys
			   )
	       )
	    ++ (fun (inode,ii,names) ->
		  let n = List.length names in
		  (** Regular files and symlinks can be
		      always deleted
		   *)
		  let ft = ii.Pfs_rpcapi_aux.filetype in
		  match ft with
		    | `ftype_regular | `ftype_symlink ->
			self # push_unlink_e ~name ~inode (n-1)
		    | `ftype_directory ->
			(** only empty directories can be deleted: *)
			self # filename_list_e ~name
			++ (fun entries ->
			      if entries <> [] then
				transerr_e `efhier esys
			      else
				self # push_unlink_e ~name ~inode (n-1)
			   )
	       ) in
	  (** We keep the locks on success. *)
	  Uq_engines.when_state
	    ~is_error:(fun _ -> unlock(); !unlock_i())
	    ~is_aborted:(fun _ -> unlock(); !unlock_i())
	    e;
	  e


  method delayed_inode_deletes =
    let inodes = ref [] in
    (** Maybe we have to unlink inode at the end of the transaction. *)
    Hashtbl.iter
      (fun inode n ->
	 if n=0 then
	   inodes := inode :: !inodes
      )
      inode_links;
    !inodes

  (** generic *)

  method pre_commit =
    (** caveat: this is not only called from Nn_commit, but also from
       Nn_coord to determine the inode notifications!
     *)
    journal

  method post_commit ok =
    self#rollback()

  method rollback () =
    (** Just release all locks *)
    shared_state # unlock_inodes (self :> < >);
    shared_state # unlock_filenames (self :> < >)
end
