(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma_client.ml 255 2010-06-24 09:35:14Z gerd $ *)

open Plasma_rpcapi_aux
open Plasma_util.Operators
open Printf

module MCSet = Set.Make (Rpc_proxy.ManagedClient)
module Coord = Plasma_rpcapi_clnt.Make'Coordination(Rpc_proxy.ManagedClient)
module Fsys = Plasma_rpcapi_clnt.Make'Filesystem(Rpc_proxy.ManagedClient)
module Dn = Plasma_rpcapi_clnt.Make'Datanode(Rpc_proxy.ManagedClient)
module Ic = Plasma_rpcapi_clnt.Make'Inode_cache(Rpc_proxy.ManagedClient)
module Util = Plasma_util

type errno = Plasma_util.errno
type strmem = [`String of string | `Memory of Netsys_mem.memory]

type buffer =
    { buf_inode : int64;
      buf_index : int64;
      buf : Netsys_mem.memory;
      mutable buf_zero : bool;    (* whether this buf should be reallocated *)
      buf_ord : int;              (* ordinal number of the buffer *)
      mutable buf_seqno : int64;  (* for cache validation only *)
      mutable buf_age : float;    (* time when the buf was last retrieved *)
      mutable buf_valid : bool;
      mutable buf_clean : bool;
    }

(** Note [buf_zero]: The flag suggests to call [Netsys_mem.zero_pages] in
    order to reallocate the buffer before the buffer memory is modified
    by direct assignment (i.e. not via [Netsys_mem.mem_read]). This flag
    is set when a virtual copy of the buffer is made, i.e. a copy with
    copy-on-write semantics.
    
    The bad sequence of operations we try to avoid:
    - Buffer is virtually copied to somewhere outside
    - Buffer is completely overwritten with a blit operation
    
    At this moment the kernel first creates a copy which is then
    immediately overwritten. Creating copy is wasted time.

    Solution: Set [buf_zero] when we know that there is a virtual copy,
    and reset if we know it is our private memory. Then, if we want
    to overwrite the whole buffer and [buf_zero] is set, we can first
    drop the copy by reallocating the memory pages.
 *)


module BLRU = 
  Plasma_util.LRU(struct 
		    type t = buffer
		    let compare b1 b2 =
		      compare b1.buf_ord b2.buf_ord
		  end)

module I64RangeSet =
  Plasma_util.RangeSet(struct
			 type t = int64
			 let compare = Int64.compare
			 let pred n =
			   if n = Int64.min_int then
			     None
			   else
			     Some(Int64.pred n)
			 let succ n =
			   if n = Int64.max_int then
			     None
			   else
			     Some(Int64.succ n)
		       end
		      )

type inodecache =
    { cache_inode : int64;
      mutable cache_access_ts : float;  (* when cache was last accessed *)
      mutable cache_val_ts : float; (* when cache was last validated *)
      mutable cache_tmo_ts : float; (* when cache must time out *)
      mutable cache_ii : inodeinfo;
      mutable cache_blocks : (int64, blockinfo list) Hashtbl.t
	(** [cache_blocks] is an incomplete mapping of indexes to
	    blockinfos
	 *)
    }


type inodemod =
    { mod_inode : int64;
      mod_bufs : (int64,buffer) Hashtbl.t;
      mutable mod_min_eof : int64;
      mutable mod_flush : unit -> unit Uq_engines.engine;
      (** The flush function is called to write dirty buffers. It should
	  update [buf_identity], [buf_block], [buf_age] but nothing else.
       *)
      mod_flush_ser :  unit Uq_engines.serializer_t;
      (** serializer for flushing all the buffers of this inode *)
      mutable mod_err : exn option
    }
    (** Stores updates of the inodeinfo performed by the background
	writers
     *)



type dn_connectivity =
    [ `Client of Rpc_proxy.ManagedClient.t
    | `Connecting of Rpc_proxy.ManagedClient.t Uq_engines.engine
    ]

type plasma_cluster =
    { name : string;
      namenodes_conf : (string * int) list;
      namenodes : Unix.sockaddr array;

      mutable down : bool;

      mutable pref_nodes : string list;
      
      mutable shm_mng : Plasma_shm.shm_manager;

      mutable buf_mem : (Netsys_mem.memory * Unix.file_descr) option;
      (** The memory used for the buffers. It is first initialized
	  when the blocksize is known. This is always shared memory
	  obtained via [mmap].
       *)

      mutable buf_next : int;
      (** The buffers 0 to [buf_next-1] are already entered into 
	  the [buffers] hashtable.
       *)

      mutable coord : Unix.sockaddr;
      (** Current coordinator *)

      mutable find_coord_ser : 
                     Rpc_proxy.ManagedClient.mclient Uq_engines.serializer_t;
      (** Serializer for the coordinator search *)

      mutable next_tid : int64;
      (** next transaction ID *)

      mutable clients : MCSet.t;
      (** Clients in use (for [close_cluster]) *)

      mutable nn_client_opt : Rpc_proxy.ManagedClient.t option;
      (** Current connection to namenode *)

      mutable dn_clients : (Unix.sockaddr, dn_connectivity) Hashtbl.t;
      (** Current connections to datanodes *)

      mutable blocksize : int;
      (** initialized first after the namenode connection is established *)

      mutable n_buffers : int;
      (** maximum number of buffers *)

      mutable act_buffers : int;
      (** actual number of buffers *)

      mutable f_buffers : int;
      (** number of buffers that are going to be flushed *)

      mutable waiting : (buffer -> unit) Queue.t;
      (** buffer users waiting for a free buffer. If this number
          is bigger than [f_buffers], a new flush operation should be started,
	  if possible.
       *)

      buffers : (int64 * int64, buffer) Hashtbl.t;
      (** buffers by (inode,index) *)

      mutable dbuffers : BLRU.t;
      (** buffers with [buf_valid && not buf_clean && not buf_flushing]. The
	  first buffer in the queue is the "oldest" (was for the longest time
	  not accessed), and is the first candidate for starting flushing
       *)

      mutable cbuffers : BLRU.t;
      (** buffers with [buf_valid && buf_clean]. First buffer is the oldest,
	  and the first candidate for reclaiming
       *)

      mutable drop_count : int64;
      (** Counts dropped buffers *)

      readers : (int64 * int64, unit Uq_engines.engine) Hashtbl.t;
      (** If a read is in progress for this block, the engine is running.
       *)

      mutable cacheclient : Rpc_proxy.ManagedClient.t option;
      (** If none, there is no inodecache. Otherwise this is the RPC client *)

      inodecache : (int64, inodecache) Hashtbl.t;
      (** Cache per inode *)

      inodecache_ser : unit Uq_engines.serializer_t;
      (** Serializer for inodecache-related queries *)

      mutable inodecache_getting_blocks :
                             (int64 * int64 * unit Uq_engines.engine) option;
      (** Whether there is a get_blocks query *)

      inodemods : (int64, inodemod) Hashtbl.t;
      (** Modification requests per inode *)

      esys : Unixqueue.event_system
    }

type plasma_trans =
    { cluster : plasma_cluster;
      tid : int64;
      mutable is_open : bool
    }

type inode = int64

type topology = [`Star|`Chain]

exception Plasma_error of errno

exception Cluster_down of string


let () =
  Netexn.register_printer (Plasma_error `einval)
    (function
       | Plasma_error code ->
	   "Plasma_error(" ^ Plasma_util.string_of_errno code ^ ")"
       | _ ->
	   assert false
    )


let cluster trans =
  trans.cluster

let cluster_namenodes c =
  c.namenodes_conf

let sync f arg =
  let e = f arg in
  Unixqueue.run e#event_system;
  match e#state with
    | `Done v -> v
    | `Error e -> raise e
    | `Aborted -> failwith "Aborted"
    | `Working _ -> assert false


let buffer_stats c =
  let out = Buffer.create 500 in
  Hashtbl.iter
    (fun (inode,index) b ->
       bprintf out "Inode %Ld block %Ld: valid=%b clean=%b \
                    in_cb=%b in_db=%b\n"
	 inode index b.buf_valid b.buf_clean
	 (BLRU.mem b c.cbuffers)
	 (BLRU.mem b c.dbuffers)
    )
    c.buffers;
  Buffer.contents out


let add_client c mc =
  c.clients <- MCSet.add mc c.clients

let del_client c mc =
  c.clients <- MCSet.remove mc c.clients

let create_nn_client ?(msg_timeout=60.0) c conn =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Plasma_rpcapi_clnt.Coordination.V1._program;
		  Plasma_rpcapi_clnt.Filesystem.V1._program;
		]
      ~msg_timeout
      ~msg_timeout_is_fatal:true
      ~initial_ping:true
      (* no idle timeout - when the connection is closed, the transactions
	 are implicitly aborted
       *)
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      c.esys in
  add_client c mclient;
  mclient

let create_ic_client c conn =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Plasma_rpcapi_clnt.Inode_cache.V1._program ]
      ~msg_timeout:1.0
      ~msg_timeout_is_fatal:true
      ~initial_ping:true
      ~idle_timeout:60.0
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      c.esys in
  add_client c mclient;
  mclient

let mfac = Hashtbl.create 5

let () =
  Hashtbl.add mfac "readdata" Xdr_mstring.string_based_mstrings


let create_dn_client c conn =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Plasma_rpcapi_clnt.Datanode.V1._program ]
      ~msg_timeout:(300.0)
      ~msg_timeout_is_fatal:true
      ~initial_ping:true
      ~idle_timeout:15.0
      ~mstring_factories:mfac
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      c.esys in
  add_client c mclient;
  mclient


let shutdown_client c mc =
  Rpc_proxy.ManagedClient.trigger_shutdown mc (fun () -> del_client c mc)


let open_cluster name nodes esys =
  if nodes = [] then invalid_arg "open_cluster";
  let namenodes =
    Array.of_list
      (List.map
	 (fun (h,p) -> Util.sockaddr_of_host_port (h,p))
	 nodes
      ) in
  { name = name;
    namenodes = namenodes;
    namenodes_conf = nodes;
    down = false;
    pref_nodes = [];
    shm_mng = new Plasma_shm.null_shm_manager();
    coord = namenodes.(0);
    find_coord_ser = Uq_engines.serializer esys;
    next_tid = 0L;
    blocksize = 0;
    clients = MCSet.empty;
    nn_client_opt = None;
    dn_clients = Hashtbl.create 5;
    buf_mem = None;
    buf_next = 0;
    n_buffers = 1;
    act_buffers = 0;
    f_buffers = 0;
    waiting = Queue.create ();
    buffers = Hashtbl.create 5;
    dbuffers = BLRU.create();
    cbuffers = BLRU.create();    
    drop_count = 0L;
    readers = Hashtbl.create 7;
    cacheclient = None;
    inodecache = Hashtbl.create 13;
    inodecache_ser = Uq_engines.serializer esys;
    inodecache_getting_blocks = None;
    inodemods = Hashtbl.create 17;
    esys = esys;
  }

let open_cluster_cc (cfg : Plasma_client_config.client_config) esys =
  let nodes =
    List.map Plasma_util.parse_host_port cfg#nn_nodes in
  open_cluster cfg#clustername nodes esys

let abort_cluster c =
  Netlog.logf `Debug "Plasma_client.abort_cluster";
  MCSet.iter
    (fun mc -> Rpc_proxy.ManagedClient.shut_down mc)
    c.clients;
  c.clients <- MCSet.empty;
  c.nn_client_opt <- None;
  Hashtbl.clear c.dn_clients

let close_cluster c =
  if not c.down then (
    c.down <- true;
    abort_cluster c;
    match c.buf_mem with
      | None -> ()
      | Some (_,fd) -> 
	  Unix.ftruncate fd 0;
	  Unix.close fd
  )

let configure_buffer c n =
  c.n_buffers <- n

let configure_pref_nodes c l =
  c.pref_nodes <- l

let configure_shm_manager c m =
  c.shm_mng <- m

let shm_manager c =
  c.shm_mng

let cluster_name c = c.name

let find_coord_e c =
  (** Find the coordinator. [coord] is our first attempt. Returns the
      new client.

      On success, tries also to create a client for the inodecache.
   *)

  let down_ic() =
    match c.cacheclient with
      | None -> ()
      | Some cc ->
	  Rpc_proxy.ManagedClient.trigger_shutdown
	    cc (fun () -> ());
	  c.cacheclient <- None in
    
  let rec search_live_nn_e addr k_next k_lim first_round =
    (** First try [addr] and submit a [find_coordinator]. If we get a
	response, everything is fine and we stop the search. Otherwise
	go on with [namenodes.(k_next)], but fail when [k=k_lim] 
     *)
    let conn = Util.connector_of_sockaddr addr in
    let mc = create_nn_client ~msg_timeout:20.0 c conn in
    Plasma_util.failsafe_e
      (Util.rpc_engine mc Coord.V1.find_coordinator'async c.name)
    ++ (fun st ->
	  shutdown_client c mc;
	  match st with
	    | `Done (Some hp) ->
		let (h,p) = Plasma_util.parse_host_port hp in
		let caddr = Plasma_util.sockaddr_of_host_port (h,p) in
		c.coord <- caddr;
		eps_e (`Done caddr) c.esys
	    | `Done None ->
		search_next_e k_next k_lim first_round
	    | `Error e ->
		Netlog.logf `Warning "find_coord_e: Exception %s" 
		  (Netexn.to_string e);
		search_next_e k_next k_lim first_round
	   | `Aborted ->
	       eps_e `Aborted c.esys
       )
  and search_next_e k_next k_lim first_round =
    if k_next = k_lim && not first_round then (
      eps_e
	(`Error (Cluster_down "find_coord_e: cannot find coordinator"))
	c.esys
    ) else (
      let addr = c.namenodes.(k_next) in
      let k' = (k_next+1) mod (Array.length c.namenodes) in
      let first_round' = first_round && k' <> k_lim in
      search_live_nn_e addr k' k_lim first_round'
    )
  in

  if c.down then
    failwith "Plasma_client: already closed";
  down_ic();
  search_live_nn_e c.coord 0 0 true
  ++ (fun addr ->
	let conn = Util.connector_of_sockaddr addr in
	let mc = create_nn_client c conn in
	Util.rpc_engine mc Coord.V1.find_inodecaches'async c.name
	++ (fun l ->
	      Netlog.logf `Debug
		"Found %d inodecache ports" (Array.length l);
	      if l <> [| |] then (
		let u = Plasma_rng.random_int (Array.length l) in
		let hp = l.(u) in
		let (h,p) = Plasma_util.parse_host_port hp in
		let conn = 
		  Plasma_util.connector_of_sockaddr
		    (Plasma_util.sockaddr_of_host_port (h,p)) in
		let ic = create_ic_client c conn in
		c.cacheclient <- Some ic
	      );
	      eps_e (`Done ()) c.esys
	   )
	>> (function
	      | `Done() ->
		  `Done mc
	      | (`Error _ | `Aborted) as st ->
		  shutdown_client c mc;
		  st
	   )
     )
    
    
let mc_info mc =
  let s =
    match Rpc_proxy.ManagedClient.mclient_state mc with
      | `Down -> "down"
      | `Connecting -> "connecting"
      | `Up None -> "up(n/a)"
      | `Up (Some addr) -> "up(" ^ Netsys.string_of_sockaddr addr ^ ")" in
  sprintf "mclient(serial=%d,state=%s)"
    (Rpc_proxy.ManagedClient.mclient_serial mc) s


let with_nn_session_e c f =
  (** Run [f mc] for a managed client [mc] of the namenode *)
  if c.down then
    failwith "Plasma_client: already closed";
  match c.nn_client_opt with
    | Some mc when Rpc_proxy.ManagedClient.mclient_state mc <> `Down ->
	f mc
    | _ ->
	c.nn_client_opt <- None;
	c.find_coord_ser # serialized
	  (fun _ ->
	     match c.nn_client_opt with
	       | Some mc ->
		   eps_e (`Done mc) c.esys
	       | _ ->
		   find_coord_e c 
		   ++ (fun mc ->
			 c.nn_client_opt <- Some mc;
			 Util.rpc_engine mc Fsys.V1.get_blocksize'async ()
			 ++ (fun size ->
			       if c.blocksize = 0 then
				 c.blocksize <- size;
			       if c.blocksize <> size then (
				 Netlog.logf `Crit
				   "After reconnecting the namenode reports a \
                                    different blocksize";
				 eps_e
				   (`Error (Failure "blocksize mismatch")) 
				   c.esys
			       )
			       else 
				 eps_e (`Done mc) c.esys
			    )
		      )
	  )
	++ (fun mc ->
	      Netlog.logf `Debug
		"with_nn_session: %s" (mc_info mc);
	      f mc
	   )


let dn_session_e c (h,p) =
  (** Create/continue a session to the datanode [hp], and start the engine
      [f mc]. Returns the engine.
   *)
  if c.down then
    failwith "Plasma_client: already closed";
  let saddr = Util.sockaddr_of_host_port (h,p) in
  let mc_conn =
    try
      Some(Hashtbl.find c.dn_clients saddr)
    with Not_found -> None in
  match mc_conn with
    | Some (`Client mc) 
	when Rpc_proxy.ManagedClient.mclient_state mc <> `Down ->
	eps_e (`Done mc) c.esys
    | Some (`Connecting e) ->
	e
    | _ ->
	Hashtbl.remove c.dn_clients saddr;
	let conn = Util.connector_of_sockaddr saddr in
	let mc = create_dn_client c conn in
	let e = 
	  Util.rpc_engine mc Dn.V1.udsocket_if_local'async ()
	  ++ (function
		| None ->
		    Hashtbl.replace c.dn_clients saddr (`Client mc);
		    Netlog.logf `Debug
		      "Datanode connection to %s:%d established" h p;
		    eps_e (`Done mc) c.esys
		| Some path ->
		    shutdown_client c mc;
		    let conn2 = Rpc_client.Unix path in
		    let mc2 = create_dn_client c conn2 in
		    Hashtbl.replace c.dn_clients saddr (`Client mc2);
		    Netlog.logf `Debug
		      "Local datanode connection to %s:%d established" h p;
		    eps_e (`Done mc2) c.esys
	     ) in
	Hashtbl.replace c.dn_clients saddr (`Connecting e);
	e


let with_new_nn_transaction_e c mc extract f =
  (** Start a new transaction on [mc], and run the engine returned by
      [f mc trans],
      and commit the transaction if this engine is successful. Otherwise
      abort it. [tid] is the transaction ID.

      The result value is extracted via [extract].
   *)
  (* TODO: option so that this transaction is repeated on ECONFLICT *)
  if c.down then
    failwith "Plasma_client: already closed";
  let tid = c.next_tid in
  c.next_tid <- Int64.succ tid;
  Netlog.logf `Debug
    "with_new_nn_transaction: tid=%Ld %s" tid (mc_info mc);
  Util.rpc_engine mc Fsys.V1.begin_transaction'async tid
  ++
    (function
       | `ok ->
	   let trans =
	     { cluster = c;
	       tid = tid;
	       is_open = true
	     } in
	   Plasma_util.failsafe_e (f mc trans)
	   ++ (fun st ->
		 trans.is_open <- false;
		 try
		   match st with
		     | `Done x ->
			 let x' = extract x in
			 new Uq_engines.map_engine
			   ~map_done:(function
					| `ok -> `Done x'
					| #errno as e -> 
					    Netlog.logf `Debug
					      "with_new_nn_transaction: \
                                                commit error tid=%Ld %s exn: %s"
					      tid (mc_info mc) 
					      (Plasma_util.string_of_errno e);
					    `Error(Plasma_error e)
				     )
			   (Util.rpc_engine 
			      mc Fsys.V1.commit_transaction'async tid)
		     | `Error err ->
			 raise err
		     | `Aborted ->
			 eps_e `Aborted c.esys
		 with
		   | err ->
		       (** Here is the risk that for abort_transaction a
			   new connection is created by ManagedClient. Nonsense.
			   Check if [mc] is already down, and do not open it
			   for only aborting.
			*)
		       Netlog.logf `Debug
			 "with_new_nn_transaction: abort tid=%Ld %s exn: %s"
			 tid (mc_info mc) (Netexn.to_string err);
		       let st1 = `Error err in
		       ( match Rpc_proxy.ManagedClient.mclient_state mc with
			   | `Down ->
			       eps_e st1 c.esys
			   | _ ->
			       new Uq_engines.map_engine
				 ~map_done:(fun _ -> st1)
				 (Util.rpc_engine
				    mc Fsys.V1.abort_transaction'async tid)
		       )
	      )
       | #errno as e ->
	    Netlog.logf `Debug
	      "with_new_nn_transaction: start error tid=%Ld %s exn: %s"
	      tid (mc_info mc) (Plasma_util.string_of_errno e);
	   eps_e (`Error (Plasma_error e)) c.esys
    )

let extract v =
  match v with
    | `ok x -> x
    | #errno as code -> raise (Plasma_error code)

let extract_unit v =
  match v with
    | `ok -> ()
    | #errno as code -> raise (Plasma_error code)

let no_extract v = v

let fextract_map f =
  function
    | `Done (`ok x) -> `Done (f x)
    | `Done (#errno as code) -> `Error(Plasma_error code)
    | `Error e -> `Error e
    | `Aborted -> `Aborted

let fextract g = fextract_map (fun x -> x) g

let fextract_map_unit x =
  function
    | `Done `ok -> `Done x
    | `Done (#errno as code) -> `Error(Plasma_error code)
    | `Error e -> `Error e
    | `Aborted -> `Aborted

let fextract_unit = fextract_map_unit ()

let plasma_error_e c code =
  eps_e (`Error (Plasma_error code)) c.esys


(* ----------------- inodecache: ------------------------ *)

(** The inodecache stores the inodeinfo record, plus a subset of the
    blockinfo records. Every access to the inodecache is validated -
    it is checked whether inodeinfo is the committed version. If so,
    the cached values can be taken.

    This assumes that the validating is_up_to_date call is very cheap.
    This is achieved by separating the inodecache server into its own
    process, and avoiding db lookups most of the time.

    The inodecache also puts not-yet-committed versions of the inodeinfo
    record into the cache. This has funny effects: First, as long as
    the commit has not happened, is_up_to_date will always return
    false, and a real get_inodeinfo lookup will happen (getting the
    version of inodeinfo that is current in the transaction). Second,
    if a parallel transaction accesses the same inode, both transactions
    will continuously overwrite the version of the competing transaction
    with their own version. Third, as soon as the transaction is committed,
    the cache stabilizes again, and already contains the right version.

    Maybe we should better avoid caching non-committed inodeinfo 
    records. However, this would require additional bookkeeping.

    There is also a cache for blocklists. We must not cache these for 
    too long, however, because these lists incluse the [node_alive]
    flag.
 *)

let ic_evict c =
  (** Limit to 20 cache entries for now *)
  if Hashtbl.length c.inodecache > 20 then (
    let oldest = ref 0L in
    let oldest_ts = ref infinity in
    Hashtbl.iter
      (fun inode ce ->
	 if ce.cache_access_ts < !oldest_ts then (
	   oldest := inode;
	   oldest_ts := ce.cache_access_ts
	 )
      )
      c.inodecache;
    Hashtbl.remove c.inodecache !oldest
  )


let get_ii_e c mc tid inode =
  (** Get inodeinfo directly from namenode *)
  Netlog.logf `Debug "get_ii_e(%Ld)" inode;
  Util.rpc_engine mc Fsys.V1.get_inodeinfo'async (tid,inode)
  ++ (fun r ->
	( match r with
	    | `ok ii ->
		if c.cacheclient <> None then (
		  (** Put this result into the cache! *)
		  let now = Unix.gettimeofday() in
		  let ce =
		    { cache_inode = inode;
		      cache_access_ts = now;
		      cache_val_ts = now;
		      cache_tmo_ts = now +. 60.0;
		      cache_ii = ii;
		      cache_blocks = Hashtbl.create 3;
		    } in
		  if not (Hashtbl.mem c.inodecache inode) then
		    ic_evict c;
		  Hashtbl.replace c.inodecache inode ce
		)
	    | _ -> ()
	);
	eps_e (`Done r) c.esys
     )

let rec get_ii_serialized_e c mc tid inode =
  let (e,signal) = Uq_engines.signal_engine c.esys in
  c.inodecache_ser # serialized
    (fun _ ->
       get_ii_e c mc tid inode
       ++ (fun r -> signal (`Done r); eps_e (`Done()) c.esys)
    )
  ++ (fun () -> e)

let get_if_cached_ii_e c mc inode =
  (** Returns inodeinfo from cache (after validation) *)
  try
    let ce = Hashtbl.find c.inodecache inode in
    ( match c.cacheclient with
	| None -> assert false
	| Some ic ->
	    (** we do not validate each time, but only if the entry
	       is not older than a certain time
	     *)
	    let now = Unix.gettimeofday() in
	    let still_valid = now < ce.cache_val_ts +. 1.0 in
	    ( if still_valid then
		(eps_e (`Done true) c.esys)
	      else (
		Netlog.logf `Debug
		  "get_if_cached_ii_e: validating inode=%Ld" inode;
		Util.rpc_engine ic Ic.V1.is_up_to_date'async (inode,ce.cache_ii)
	      )
	    )
	    ++ (fun flag ->
		  if flag && now < ce.cache_tmo_ts then (
		    (*Netlog.logf `Debug "get_if_cached_ii_e(%Ld): from cache"
		      inode;
		     *)
		    ce.cache_access_ts <- now;
		    if not still_valid then
		      ce.cache_val_ts <- now;
		    eps_e (`Done (Some ce.cache_ii)) c.esys
		  )
		  else (
		    Hashtbl.remove c.inodecache inode;
		    eps_e (`Done None) c.esys
		  )
	       )
    )
  with
    | Not_found ->
	eps_e (`Done None) c.esys


let get_cached_ii_e c mc tid inode =
  (** Returns inodeinfo from cache (after validation), or loads it *)
  get_if_cached_ii_e c mc inode
  ++ (function
	| Some ii ->
	    eps_e (`Done(`ok ii)) c.esys
	| None ->
	    get_ii_serialized_e c mc tid inode
     )


let check_up_to_date_e c mc inode seqno =
  (** Returns whether this seqno is the seqno of the inode *)
  match c.cacheclient with
    | None -> assert false
    | Some ic ->
	Util.rpc_engine ic Ic.V1.is_up_to_date_seqno'async (inode,seqno)
	++ (function
	      | true ->
		  eps_e (`Done true) c.esys
	      | false ->
		  with_new_nn_transaction_e c mc (fun x -> x)
		    (fun mc trans ->
		       get_ii_serialized_e c mc trans.tid inode
		       ++ (fun r ->
			     let flag =
			       match r with 
				 | `ok ii -> ii.seqno = seqno
				 | _ -> false in
			     eps_e (`Done flag) c.esys
			  )
		    )
	   )


let get_blocks_e c mc tid_opt inode index number evict =
  ( match tid_opt with
      | None ->
	  (** Case: no transaction yet. Create one on the fly. *)
	  with_new_nn_transaction_e c mc (fun x -> x)
	    (fun mc trans ->
	       Plasma_util.rpc_engine
		 mc Fsys.V1.get_blocks'async(trans.tid,inode,index,number)
	    )
      | Some tid ->
	  (** Case: we have already a transaction. Use this. *)
	  Plasma_util.rpc_engine
	    mc Fsys.V1.get_blocks'async(tid,inode,index,number)
  )
  ++ (fun r ->
	( match r with
	    | `ok bl ->
		Netlog.logf `Debug
		  "get_blocks_e: inode=%Ld got %d blocks" 
		  inode (Array.length bl);
		( try
		    let ce = Hashtbl.find c.inodecache inode in
		    (** Put this result into the cache! *)
		    if evict then Hashtbl.clear ce.cache_blocks;
		    Array.iter
		      (fun block ->
			 let l =
			   try Hashtbl.find ce.cache_blocks block.index
			   with Not_found -> [] in
			 Hashtbl.replace ce.cache_blocks block.index (block::l)
		      )
		      bl
		  with
		    | Not_found -> ()
		)
	    | _ -> ()
	);
	eps_e (`Done r) c.esys
     )
	  
let get_blocks_serialized_e c mc tid_opt inode index number evict =
  let (e,signal) = Uq_engines.signal_engine c.esys in
  let s =
    c.inodecache_ser # serialized
      (fun _ ->
	 get_blocks_e c mc tid_opt inode index number evict
	 >> (fun st ->
	       c.inodecache_getting_blocks <- None;
	       signal st;
	       `Done ()
	    )
      ) in
  c.inodecache_getting_blocks <- Some(index,number,s);
  s ++ (fun () -> e)

let get_several_blocks_e c mc tid_opt inode index =
  (** Fill the cache with some blocks, but return only [index] *)
  get_blocks_serialized_e c mc tid_opt inode index 100L true
  ++ (fun r ->
	match r with
	  | `ok bl ->
	      (** Extract only the requested index: *)
	      let bl' =
		Array.of_list
		  (List.fold_left
		     (fun acc block ->
			if block.index = index then
			  block :: acc
			else
			  acc
		     )
		     []
		     (Array.to_list bl)
		  ) in
	      eps_e (`Done (`ok bl')) c.esys
	  | #errno as errno ->
	      eps_e (`Done errno) c.esys
     )

let rec get_cached_blocks_e c mc tid_opt inode index =
  (** Get the blocks for this index from cache (after validation).
      The caller has to revalidate after reading the blocks!
   *)
  get_if_cached_ii_e c mc inode
  ++ (fun ii_opt ->
	try
	  (** At this point the cache is validated - if that was necessary *)
	  let ce = Hashtbl.find c.inodecache inode in
	  ( try
	      let blocks = Hashtbl.find ce.cache_blocks index in
	      (** Respond from cache: *)
	      Netlog.logf `Debug
		"get_cached_blocks inode=%Ld index=%Ld: from cache"
		inode index;
	      eps_e (`Done (`ok (Array.of_list blocks))) c.esys
	    with
	      | Not_found ->
		  (** Not in cache: there might already be a get_blocks
		      query. Check that first
		   *)
		  let reaction =
		    match c.inodecache_getting_blocks with
		      | None ->
			  `Request_several_blocks
		      | Some(i,l,e) ->
			  if index >= i && index < Int64.add i l then
			    `Wait_for_request e
			  else
			    `Request_single_block in
		  match reaction with
		    | `Request_single_block ->
			raise Not_found (* handled below *)
		    | `Request_several_blocks ->
			get_several_blocks_e c mc tid_opt inode index
		    | `Wait_for_request e ->
			(** Wait until the current request is done, and
			   then check the cache again.
			 *)
			Netlog.logf `Debug
			  "get_cached_blocks inode=%Ld index=%Ld: waiting"
			  inode index;
			e
			++ (fun _ ->
			      get_cached_blocks_e c mc tid_opt inode index
			   )
	  )
	with
	  | Not_found ->
	      (** Case: we do not have a cache *)
	      Netlog.logf `Debug
		"get_cached_blocks inode=%Ld index=%Ld: no cache"
		inode index;
	      get_blocks_serialized_e c mc tid_opt inode index 1L false
     )

(* ---- transaction statements ---- *)

let start_e c =
  with_nn_session_e c
    (fun mc ->
       let tid = c.next_tid in
       c.next_tid <- Int64.succ tid;
       let trans =
	 { cluster = c;
	   tid = tid;
	   is_open = true
	 } in
       Util.rpc_engine mc Fsys.V1.begin_transaction'async tid
       ++ (function
	     | `ok ->
		 eps_e (`Done trans) c.esys
	     | #errno as code ->
		 plasma_error_e c code
	  )
    )

let start trans = sync start_e trans

let abort_e trans =
  if not trans.is_open then
    plasma_error_e trans.cluster `enotrans
  else (
    trans.is_open <- false;
    with_nn_session_e trans.cluster
      (fun mc ->
	 Util.rpc_engine mc Fsys.V1.abort_transaction'async trans.tid
	 >> fextract_unit
      )
  )

let abort trans = sync abort_e trans

let commit_e trans =
  if not trans.is_open then
    plasma_error_e trans.cluster `enotrans
  else (
    trans.is_open <- false;
    with_nn_session_e trans.cluster
      (fun mc ->
	 Util.rpc_engine mc Fsys.V1.commit_transaction'async trans.tid
	 >> fextract_unit
      )
  )

let commit trans = sync commit_e trans

let blocksize_e c =
  if c.blocksize <> 0 then
    eps_e (`Done c.blocksize) c.esys
  else (
    with_nn_session_e c
      (fun mc ->
	 eps_e (`Done c.blocksize) c.esys
      )
  )

let blocksize c = sync blocksize_e c

let fsstat_e c =
  with_nn_session_e c
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.get_fsstat'async ()
       >> fextract
    )

let fsstat c = sync fsstat_e c


let local_identities_e c =
  with_nn_session_e c
    (fun mc ->
       Util.rpc_engine mc Coord.V1.local_identities'async c.name
       ++ (fun l -> eps_e (`Done(Array.to_list l)) c.esys)
    )

let local_identities c = sync local_identities_e c


let get_inodeinfo_e trans inode =
  with_nn_session_e trans.cluster
    (fun mc ->
       (* Util.rpc_engine mc Fsys.V1.get_inodeinfo'async (trans.tid,inode) *)
       get_cached_ii_e trans.cluster mc trans.tid inode
       >> fextract
    )

let get_inodeinfo trans inode = sync(get_inodeinfo_e trans) inode


let set_inodeinfo_e trans inode ii =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.update_inodeinfo'async (trans.tid,inode,ii)
       >> fextract_unit
    )

let set_inodeinfo trans inode ii = sync(set_inodeinfo_e trans inode) ii


let get_blocklist_e trans inode block n =
  let n = Int64.of_int n in
  with_nn_session_e trans.cluster
    (fun mc ->
       ( Util.rpc_engine mc Fsys.V1.get_blocks'async (trans.tid,inode,block,n)
	 >> fextract
       ) ++ (fun l -> eps_e (`Done(Array.to_list l)) trans.cluster.esys)
    )

let get_blocklist trans inode block n = 
  sync(get_blocklist_e trans inode block) n


let create_inode_e trans ii =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.allocate_inode'async (trans.tid,ii)
       >> fextract
    )

let create_inode trans ii = sync(create_inode_e trans) ii


let delete_inode_e trans inode =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.delete_inode'async (trans.tid,inode)
       >> fextract_unit
    )

let delete_inode trans inode = sync (delete_inode_e trans) inode


let lookup_e trans name =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.lookup'async (trans.tid,name,(-1L))
       >> fextract
    )

let lookup trans name = sync (lookup_e trans) name

let dir_lookup_e trans dir_inode name =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.lookup'async (trans.tid,name,dir_inode)
       >> fextract
    )

let dir_lookup trans dir_inode name = sync (dir_lookup_e trans dir_inode) name

let rev_lookup_e trans inode =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.rev_lookup'async (trans.tid,inode)
       >> fextract_map Array.to_list
    )

let rev_lookup trans inode = sync (rev_lookup_e trans) inode

let link_count_e trans inode =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.link_count'async (trans.tid,inode)
       >> fextract
    )

let link_count trans inode = sync (link_count_e trans) inode

let link_e trans name inode =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.link'async (trans.tid,name,inode)
       >> fextract_unit
    )

let link trans name inode = sync (link_e trans name) inode

let unlink_e trans name =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.unlink'async (trans.tid,name)
       >> fextract_unit
    )

let unlink trans name = sync (unlink_e trans) name

let unwrap_entries entries =
  Array.map
    (fun ent ->
       (ent.entry_name, ent.entry_inode)
    )
    entries

let list_inode_e trans inode =
  with_nn_session_e trans.cluster
    (fun mc ->
       Util.rpc_engine mc Fsys.V1.list'async (trans.tid,inode)
       >> fextract_map (fun x -> Array.to_list(unwrap_entries x))
    )


let list_inode trans inode = sync (list_inode_e trans) inode

let list_e trans name =
  with_nn_session_e trans.cluster
    (fun mc ->
       ( Util.rpc_engine mc Fsys.V1.lookup'async (trans.tid,name,(-1L))
	 >> fextract
       ) 
       ++ (fun inode ->
	     Util.rpc_engine mc Fsys.V1.list'async (trans.tid,inode)
	     >> fextract_map (fun x -> Array.to_list(unwrap_entries x))
	  )
    )

let list trans name = sync (list_e trans) name

let create_file_e trans name ii =
  if ii.filetype = `ftype_directory then
    invalid_arg "Plasma_client.create_file";
  with_nn_session_e trans.cluster
    (fun mc ->
       ( Util.rpc_engine mc Fsys.V1.allocate_inode'async (trans.tid,ii)
	 >> fextract
       )
       ++ (fun inode ->
	     Util.rpc_engine mc Fsys.V1.link'async (trans.tid,name,inode)
	     >> fextract_map_unit inode
	  )
    )

let create_file trans name ii = sync (create_file_e trans name) ii

let mkdir_e trans name ii =
  if ii.filetype <> `ftype_directory then
    invalid_arg "Plasma_client.mkdir";
  with_nn_session_e trans.cluster
    (fun mc ->
       ( Util.rpc_engine mc Fsys.V1.allocate_inode'async (trans.tid,ii)
	 >> fextract
       )
       ++ (fun inode ->
	     Util.rpc_engine mc Fsys.V1.link'async (trans.tid,name,inode)
	     >> fextract_map_unit inode
	  )
    )

let mkdir trans name ii = sync (mkdir_e trans name) ii


let create_ii ftype mode field1 =
  let t = Unix.gettimeofday() in
  let t1 = Int64.of_float t in
  let t2 = truncate ((t -. Int64.to_float t1) *. 1E9) in
  { filetype = ftype;
    usergroup = { user = "foo"; group = "bar" };  (* TODO *)
    mode = mode;
    eof = 0L;
    replication = 0;  (* means: server takes default *)
    blocklimit = 0L;
    mtime = { tsecs = t1; tnsecs = t2 };
    ctime = { tsecs = t1; tnsecs = t2 };
    field1 = field1;
    seqno = 0L;
    create_verifier = 0L;
  }
 
let regular_ii c mode =
  create_ii `ftype_regular (mode land 0o755) ""

(* TODO: use process umask. Problem: Unix.umask does not allow to only
   query the umask
 *)

let symlink_ii c dest =
  create_ii `ftype_symlink 0o777 dest

let dir_ii c mode =
  create_ii `ftype_directory (mode land 0o755) ""



(* ----------------- buffered access: ------------------- *)

let get_mem c blocksize =
  if c.down then
    failwith "Plasma_client: already closed";
  match c.buf_mem with
    | Some (m,fd) -> (m,fd)
    | None ->
	let ps = Netsys_mem.getpagesize() in
	if blocksize mod ps <> 0 then
	  failwith "Blocksize is not a multiple of the page size";
	let (fd,name) = Plasma_util.get_shm_fd "plasmaclient" 0 in
	try
	  Netsys_posix.shm_unlink name;
	  let mem_size = blocksize * c.n_buffers in
	  Plasma_util.enlarge_shm fd mem_size;
	  c.shm_mng # book_allocation (c.shm_mng # new_id()) mem_size;
	  let mem = Plasma_util.map_shm fd 0L mem_size in
	  c.buf_mem <- Some (mem,fd);
	  c.act_buffers <- c.n_buffers;
	  (mem,fd)
	with
	  | error ->
	      Unix.close fd;
	      raise error


let new_buffer c inode index blocksize =
  (** Allocate a new invalid buffer record for [(inode,index)] *)
  let (m,_) = get_mem c blocksize in
  let p = c.buf_next in
  let pos = p * blocksize in
  let b =
    { buf_inode = inode;
      buf_index = index;
      buf_seqno = 0L;
      buf_age = 0.0;
      buf_valid = false;
      buf_clean = false;
      buf_ord = p;
      buf_zero = false;
      buf = Bigarray.Array1.sub m pos blocksize;
    } in
  c.buf_next <- c.buf_next + 1;
  Hashtbl.add c.buffers (inode,index) b;
  b

let zero_buffer b =
(*
  let blocksize = Bigarray.Array1.dim b.buf in
  try Netsys_mem.zero_pages b.buf 0 blocksize
  with Invalid_argument _ -> ()
 *)
  ()   (* observing worse performance when enabled... *)

let blit_from_buffer c b bpos sm spos len =
  match sm with
    | `String s ->
	Netsys_mem.blit_memory_to_string b.buf bpos s spos len
    | `Memory m ->
	(* optimize whole buffer copies: *)
	let blocksize = Bigarray.Array1.dim b.buf in
	if len = blocksize && bpos = 0 && c.buf_mem <> None then
	  let (_,fd) = get_mem c blocksize in
	  ignore(Unix.lseek fd (b.buf_ord * blocksize) Unix.SEEK_SET);
	  ignore(Netsys_mem.mem_read fd m bpos len);
	  b.buf_zero <- true
	else
	  Bigarray.Array1.blit
	    (Bigarray.Array1.sub b.buf bpos len)
	    (Bigarray.Array1.sub m spos len)


let blit_to_buffer c sm spos b bpos len =
  let blocksize = Bigarray.Array1.dim b.buf in
  match sm with
    | `String s ->
	if b.buf_zero && bpos = 0 && len = blocksize then
	  zero_buffer b;
	Netsys_mem.blit_string_to_memory s spos b.buf bpos len;
	b.buf_zero <- false;
    | `Memory m ->
	(* optimize whole buffer copies: *)
	if len = blocksize && bpos = 0 && c.buf_mem <> None then
	  let (_,fd) = get_mem c blocksize in
	  ignore(Unix.lseek fd (b.buf_ord * blocksize) Unix.SEEK_SET);
	  ignore(Netsys_mem.mem_write fd m bpos len);
	  b.buf_zero <- true;
	else (
	  Bigarray.Array1.blit
	    (Bigarray.Array1.sub m spos len)
	    (Bigarray.Array1.sub b.buf bpos len);
	  b.buf_zero <- false;
	)
	    

let reuse_clean_buffer c inode index =
  (** Try to find a clean buffer record, and turn it into an invalid
      record for [(inode,index)]. If there is no clean buffer raise
      [Not_found]
   *)
  let b0 = BLRU.take c.cbuffers in (* may raise Not_found *)
  Hashtbl.remove c.buffers (b0.buf_inode,b0.buf_index);
  let b = { b0 with 
	      buf_inode = inode;
	      buf_index = index;
	      buf_valid = false;
	  } in
  Hashtbl.add c.buffers (inode,index) b;
  b


let remove_from_cbuffers c b0 =
  (** Remove b0 from [cbuffers] *)
  BLRU.remove b0 c.cbuffers


let remove_from_dbuffers c b0 =
  (** Remove b0 from [dbuffers] *)
  BLRU.remove b0 c.dbuffers


let reuse_this_buffer c inode index b0 =
  (** Turn the buffer [b] into an invalid buffer for
      [(inode,index)].
   *)
  if b0.buf_valid && b0.buf_clean then
    remove_from_cbuffers c b0;
  if b0.buf_valid && not b0.buf_clean then
    remove_from_dbuffers c b0;
  Hashtbl.remove c.buffers (b0.buf_inode,b0.buf_index);
  let b = { b0 with 
	      buf_inode = inode;
	      buf_index = index;
	      buf_valid = false;
	  } in
  Hashtbl.add c.buffers (inode,index) b;
  b


let oldify_clean_buffer c b =
  (** Enforce that a clean buffer [b] is moved to the end of [cbuffers] *)
  assert(b.buf_valid && b.buf_clean);
  BLRU.add b c.cbuffers


let set_clean_from_invalid b ii =
  (** Turn an invalid buffer into a clean buffer. It is not added to [cbuffers]
   *)
  assert(not b.buf_valid);
  b.buf_seqno <- ii.seqno;
  b.buf_age <- Unix.gettimeofday();
  b.buf_valid <- true;
  b.buf_clean <- true
  

let set_clean_from_dirty c b =
  (** Turn a dirty buffer into a clean buffer. It is not added to [cbuffers] *)
  assert(b.buf_valid && not b.buf_clean);
  b.buf_clean <- true;
  remove_from_dbuffers c b


let invalidate_buffer c b =
  (** If b is valid turn it into invalid *)
  if b.buf_valid then (
    if not b.buf_clean  then
      Netlog.logf `Crit
	"Invalidating dirty buffer. This should never happen! (data loss)";
    if b.buf_clean then
      remove_from_cbuffers c b;
    if not b.buf_clean then
      remove_from_dbuffers c b;
    b.buf_valid <- false;
  )


let drop_buffer c b0 =
  (** We don't have a real state for this case - unused buffer.
      For simplicity we just claim this buffer as clean for a
      non-existing inode number 

      FIXME: represent this in a cleaner way
   *)
  if b0.buf_valid && b0.buf_clean then
    remove_from_cbuffers c b0;
  if b0.buf_valid && not b0.buf_clean then
    remove_from_dbuffers c b0;
  Hashtbl.remove c.buffers (b0.buf_inode,b0.buf_index);
  let index = c.drop_count in
  c.drop_count <- Int64.succ index;
  let b = { b0 with 
	      buf_inode = (-1L);
	      buf_index = index;
	      buf_valid = true;
	  } in
  Hashtbl.add c.buffers (-1L,index) b;
  if not (BLRU.mem b c.cbuffers) then
    BLRU.add b c.cbuffers; (* non-optimal - better add in front position *)
  b


let rec flush_inode_e c inode =
  (* flush all dirty buffers of an inode *)
  try
    let m = Hashtbl.find c.inodemods inode in
    m.mod_flush_ser # serialized
      (fun _ ->
	 (** Be careful here - [m] could already be removed here, or be
	     replaced
	  *)
	 try
	   let m' = Hashtbl.find c.inodemods inode in
	   if m' == m then
	     m.mod_flush()
	   else
	     flush_inode_e c inode
	 with
	   | Not_found -> 
	       eps_e (`Done()) c.esys
      )
  with
    | Not_found ->
	eps_e (`Done()) c.esys


let maybe_flush_buffer c =
  (** Only flush a buffer if there are users that can benefit: *)
  let do_it = Queue.length c.waiting > c.f_buffers in
  Netlog.logf `Debug "Plasma_client.maybe_flush_buffer: \
                      #waiting=%d f_buffer=%d do_it=%b"
    (Queue.length c.waiting) c.f_buffers do_it;
  if do_it then (
    (** Check whether there are buffers we can flush: *)
    try
      let b = BLRU.take c.dbuffers in
      let inode = b.buf_inode in
      ignore(flush_inode_e c inode)
    with
      | Not_found ->
	  ()
  )


let find_buffer_e c blocksize inode index =
  (** Find this buffer, or request a new buffer. The returned buffer is
      always invalid.

      It can also happen that we return [None]: this means a concurrent
      access was faster, and got the buffer for (inode,index).

      FIXME: we should return None asap, and not only when the request
      is seen in c.waiting
   *)
  ignore(get_mem c blocksize);  (* ensure that [c.act_buffers] is set *)
  try
    let b = Hashtbl.find c.buffers (inode,index) in
    Netlog.logf `Debug
      "find_buffer_e inode=%Ld index=%Ld: reusing block buffer" inode index;
    invalidate_buffer c b;
    eps_e (`Done(Some b)) c.esys
  with
    | Not_found ->
	(** If we have less than [act_buffers], just create a new one *)
	if Hashtbl.length c.buffers < c.act_buffers then (
	  Netlog.logf `Debug
	    "find_buffer_e inode=%Ld index=%Ld: creating buffer" inode index;
	  let b = new_buffer c inode index blocksize in
	  eps_e (`Done(Some b)) c.esys
	)
	else (
	  (** If there is a clean buffer, take it *)
	  try
	    let b = reuse_clean_buffer c inode index in
	    Netlog.logf `Debug
	      "find_buffer_e inode=%Ld index=%Ld: reusing clean buffer"
	      inode index;
	    eps_e (`Done(Some b)) c.esys
	  with
	    | Not_found ->
		(** Add ourselves to the list of waiting users: *)
		Netlog.logf `Debug
		  "find_buffer_e inode=%Ld index=%Ld: waiting for buffer"
		  inode index;
		let (e,signal) = Plasma_util.signal_engine c.esys in
		let on_buffer b0 = 
		  if Hashtbl.mem c.buffers (inode,index) then (
		    Netlog.logf `Debug
		      "find_buffer_e inode=%Ld index=%Ld: buffer already found"
		      inode index;
		    signal(`Done None)
		  ) else (
		    Netlog.logf `Debug
		      "find_buffer_e inode=%Ld index=%Ld: got buffer"
		      inode index;
		    let b = reuse_this_buffer c inode index b0 in
		    signal(`Done(Some b))
		  ) in
		Queue.add on_buffer c.waiting;
		(** Check whether another buffer should be flushed: *)
		maybe_flush_buffer c;
		(** Wait until we are notified: *)
		e
	)


let all_dead blocklist =
  let dead = ref true in
  Array.iter (fun block -> if block.node_alive then dead := false) blocklist;
  !dead


let post_action c b =
  (** If the pressure is high, we may have to reuse
      this buffer immediately:
   *)
  if b.buf_clean then (
    try
      let w = Queue.take c.waiting in
      w b
    with
      | Queue.Empty ->
	  oldify_clean_buffer c b
  )
    


let rec really_rdblock_e c mc tid_opt ii blocksize inode index b action =
  assert(not b.buf_valid);
  Netlog.logf `Debug
    "really_rdblock_e inode=%Ld index=%Ld" inode index;

  let call_action_e () =
    (** Called once [b.buf] is filled with data *)
    set_clean_from_invalid b ii;
    action b;  (** may clear buf_clean! *)
    Hashtbl.remove c.readers (inode,index);
    post_action c b;
    eps_e (`Done ()) c.esys
  in

  (** Get the locations of the block: *)
  get_cached_blocks_e c mc tid_opt inode index
  ++ (function
	| `ok blocklist ->
	    Netlog.logf `Debug
	      "really_rdblock_e inode=%Ld index=%Ld: got blocklist n=%d"
	      inode index (Array.length blocklist);

	    (** If the blocklist is non-empty but all nodes are dead,
		return an error
	     *)
	    if blocklist <> [| |] && all_dead blocklist then
	      plasma_error_e c `eio
	    else
	      (** If the blocklist is empty, we are reading a file hole.
		  Return zeros to the user
	       *)
	      if blocklist = [| |] then (
		Bigarray.Array1.fill b.buf '\000';
		call_action_e ()
	      )
	      else (
		(** Get any block: *)
		(* Better: Instead of taking a block randomly, look at
		   the number of outstanding responses.
		 *)
		let j = Plasma_rng.random_int (Array.length blocklist) in
		let block = blocklist.(j) in  
		let hp = Util.parse_host_port block.node in
		Netlog.logf `Debug
		  "really_rdblock_e inode=%Ld index=%Ld: reading %Ld from %s"
		  inode index block.block block.node;
		dn_session_e c hp
		++ (fun dc ->
		      (* TODO: optimize local accesses with `dnch_shm *)
		      Util.rpc_engine 
			dc Dn.V1.read'async 
			(`dnch_rpc, block.block, 0, blocksize, 
			 block.safetrans_id, block.safetrans_vfy)
		      ++ (fun dnch ->
			    let data =
			      match dnch with
				| `dnch_rpc data -> data
				| `dnch_shm -> assert false in
			    assert(data#length = blocksize);
			    (** Now it is possible that the block has been
				released in the meantime. We validate again
				because of this
			     *)
			    Netlog.logf `Debug
			      "really_rdblock_e inode=%Ld index=%Ld: \
                                got %Ld from %s"
			      inode index block.block block.node;
			    check_up_to_date_e c mc inode block.inode_seqno
			      (* eps_e (`Done true) c.esys *)
			    ++ (fun ok ->
				  if ok then (
				    Netlog.logf `Debug
				      "really_rdblock_e inode=%Ld index=%Ld: \
                                         returning block to user"
				      inode index;
				    (** Reallocate the buffer page if this is
					supported by the OS and if we think
					this is good for copy-on-write
				     *)
				    if b.buf_zero then
				      zero_buffer b;
				    (* This blit is a string-to-bigarray copy *)
				    data # blit_to_memory 0 b.buf 0 blocksize;
				    call_action_e()
				  )
				  else (
				    Netlog.logf `Debug
				      "really_rdblock_e inode=%Ld index=%Ld: \
                                         block is outdated, retrying"
				      inode index;
				    really_rdblock_e c mc tid_opt ii 
				      blocksize inode index b action
				  )
			       )
			 )
		  )
	      )
	| #errno as e ->
	    plasma_error_e c e
     )
    
let rec rdblock_e c mc tid_opt ii blocksize inode index action =
  (** Read this block and perform [action] *)
  Netlog.logf `Debug
    "rdblock_e inode=%Ld index=%Ld" inode index;
  try
    let b = Hashtbl.find c.buffers (inode,index) in
    if not b.buf_valid then raise Not_found;
    (** Cache validity check: *)
    if b.buf_clean && ii.seqno <> b.buf_seqno then (
      (** invalid *)
      invalidate_buffer c b;
      raise Not_found
    );
    (** If the buffer is dirty and the seqno is wrong, we have the case
	of competing writes. We cannot do anything then.
     *)
    action b;
    post_action c b;
    eps_e (`Done ()) c.esys
  with
    | Not_found ->
	(** Maybe somebody is already reading this block. In this case
	    we just attach to this engine.
	 *)
	try
	  let e = Hashtbl.find c.readers (inode,index) in
	  Netlog.logf `Debug
	    "rdblock_e inode=%Ld index=%Ld: attaching to existing read"
	    inode index;
	  e ++ (fun () -> 
		  rdblock_e c mc tid_opt ii blocksize inode index action)
	with
	  | Not_found ->
	      (** We don't have a valid copy of this block. Get a buffer,
		  and start reading.
	       *)
	      Netlog.logf `Debug
		"rdblock_e inode=%Ld index=%Ld: finding buffer" inode index;
	      find_buffer_e c blocksize inode index
	      ++ (fun b_opt ->
		    match b_opt with
		      | None ->
			  (** Some write jumped in in the meantime: *)
			  rdblock_e c mc tid_opt ii blocksize inode index action
		      | Some b ->
			  let e = 
			    really_rdblock_e
			      c mc tid_opt ii blocksize inode index b action in
			  Hashtbl.add c.readers (inode,index) e;
			  e
		 )

let read_with_ii_e c mc tid_opt inode ii pos sm spos len =
  let blocksize = c.blocksize in 
  let blocksizeL = Int64.of_int blocksize in
  let max_len = (** how much we could read at most *)
    max (Int64.sub ii.eof pos) 0L in
  let eff_len = (** how much we will read into [sm] *)
    Int64.to_int(min max_len (Int64.of_int len)) in
  let end_pos =
    Int64.add pos (Int64.of_int eff_len) in
  let at_eof =
    end_pos >= ii.eof in

  let rec submit_read_requests offs =
    (** create read requests for [pos+offs] on *)
    let offsL = Int64.of_int offs in
    let cur_pos = Int64.add pos offsL in
    let index =
      Int64.div cur_pos blocksizeL in
    let bpos = 
      Int64.to_int (Int64.sub cur_pos (Int64.mul index blocksizeL)) in
    let req_len =
      min (eff_len - offs) (blocksize - bpos) in
    if req_len = 0 then
      []
    else
      let e =
	rdblock_e c mc tid_opt ii blocksize inode index 
	  (fun b ->
	     blit_from_buffer c b bpos sm (spos+offs) req_len
	  ) in
      e :: submit_read_requests (offs + req_len)
  in
  
  Uq_engines.msync_engine
    (submit_read_requests 0)
    (fun () () -> ())
    ()
    c.esys
  ++
    (fun () ->
	  eps_e (`Done(eff_len,at_eof,ii)) c.esys
    )


let sm_length =
  function
    | `String s -> String.length s 
    | `Memory m -> Bigarray.Array1.dim m
    

let read_trans_e trans inode pos sm spos len =
  let c = trans.cluster in
  let l = sm_length sm in
  if spos < 0 || len < 0 || spos > l-len then
    invalid_arg "Plasma_client.read_trans_e";

  let get_info_e mc =
    (** Get information: inodeinfo *)
    get_cached_ii_e c mc trans.tid inode
    >> fextract in

  with_nn_session_e c
    (fun mc ->
       get_info_e mc
       ++ (fun ii ->
	     read_with_ii_e 
	       c mc (Some trans.tid) inode ii pos sm spos len 
	  )
    )

let read_e c inode pos sm spos len =
  let l = sm_length sm in
  if spos < 0 || len < 0 || spos > l-len then
    invalid_arg "Plasma_client.read_e";

  with_nn_session_e c
    (fun mc ->
       get_if_cached_ii_e c mc inode
       ++ (function
	     | Some ii ->
		 (** Luck. We have a recently validated inodeinfo *)
		 read_with_ii_e
		   c mc None inode ii pos sm spos len
	     | None ->
		 (** Start a transaction, and call read_e *)
		 with_new_nn_transaction_e c mc (fun x -> x)
		   (fun mc trans ->
		      read_trans_e trans inode pos sm spos len
		   )
	  )
    )

let read c inode pos sm spos len =
  sync (read_e c inode pos sm spos) len


let pause =
  [| 0.01; 0.02; 0.04; 0.08; 0.16; 0.32; 0.64; 1.0;
     2.0; 4.0; 8.0; 16.0; 32.0; 60.0
  |]

let post_wrbuffer c b =
  (** If we have waiting users, immediately forward.
      Otherwise set the state of the buffer to "clean"
   *)
  Netlog.logf `Debug
    "Plasma_client.post_wrbuffer: inode=%Ld #waiting=%d" 
    b.buf_inode (Queue.length c.waiting);
  try
    let w = Queue.take c.waiting in
    w b
  with
    | Queue.Empty ->
	set_clean_from_dirty c b;
	oldify_clean_buffer c b


let wrbuffer_e c inodemod b blocklist seqno n_success =
  (** Writes buffer b out *)
  let blocksize = Bigarray.Array1.dim b.buf in

  let write_block_e block =
    let ms =
      Xdr_mstring.memory_based_mstrings # create_from_memory
	b.buf
	0
	blocksize
	false in
    b.buf_zero <- true;
    let hp = Plasma_util.parse_host_port block.node in
    (* TODO: use dnch_shm for local writes *)
    dn_session_e c hp
    ++ (fun dc ->
	  Plasma_util.rpc_engine dc Dn.V1.write'async
	    (block.block, `dnch_rpc ms, block.safetrans_id, block.safetrans_vfy)
       )
  in
  
  let rec write_blocks_e blocklist =
    match blocklist with
      | [] ->
	  (** We are done with the buffer. *)
	  b.buf_seqno <- seqno;
	  b.buf_age <- Unix.gettimeofday();
	  Hashtbl.remove inodemod.mod_bufs b.buf_index;
	  if Hashtbl.length inodemod.mod_bufs = 0 then
	    Hashtbl.remove c.inodemods inodemod.mod_inode;
	  c.f_buffers <- c.f_buffers - 1;
	  incr n_success;
	  post_wrbuffer c b;
	  eps_e (`Done ()) c.esys
      | block :: blocklist' ->
	  write_block_e block
	  ++
	    (fun _ -> write_blocks_e blocklist')
  in

  write_blocks_e blocklist
  

let wrbuffers_e c inodemod imin imax blocklist seqno n_success =
  let blocks_of_index = Hashtbl.create 5 in
  
  Array.iter
    (fun block ->
       let blocks =
	 try Hashtbl.find blocks_of_index block.index
	 with Not_found -> [] in
       Hashtbl.replace blocks_of_index block.index (block :: blocks)
    )
    blocklist;
  
  let rec write_loop i =
    if i > imax then
      eps_e (`Done ()) c.esys
    else (
      try
	  let blocks = Hashtbl.find blocks_of_index i in
	  let b = Hashtbl.find inodemod.mod_bufs i in
	  wrbuffer_e c inodemod b blocks seqno n_success
	  ++ (fun () -> write_loop (Int64.succ i))
      with
	| Not_found -> assert false
    )
  in
  
  write_loop imin
    


let flinode_e c inodemod =
  (** Flushes the dirty blocks of an inode. 
      (Note that this does not set [buf_clean] at the
      end, this is left to the caller.)

      As flushing happens in a separate transaction, we can face the
      following problems:
       - the inode does not exist anymore
       - the blocks to write are behind EOF
   *)
  let inode = inodemod.mod_inode in
  let min_eof = inodemod.mod_min_eof in

  let n = Hashtbl.length inodemod.mod_bufs in  (* total buffers *)
  let n_success = ref 0 in (* successful buffers so far *)

  let ranges =
    Hashtbl.fold
      (fun index b acc ->
	 assert(b.buf_valid && not b.buf_clean);
	 I64RangeSet.add index acc
      )
      inodemod.mod_bufs
      I64RangeSet.empty in

  let rangelist = ref [] in
  I64RangeSet.iter
    (fun imin imax -> rangelist := (imin,imax) :: !rangelist)
    ranges;
  rangelist := List.rev !rangelist;

  (** Note that once we start sub-engines, inodemod can be modified by
      concurrent activities.
   *)
  
  let cur_trans = ref None in

  let rec trans_repeat mc rep f =
    if rep > 10 then (   (* TODO: this limit should be configurable *)
      Netlog.logf `Crit
	"Stopping flush after %d write errors" rep;
      plasma_error_e c `efailed
    )
    else (
      ( match !cur_trans with
	  | None -> start_e c
	  | Some t -> eps_e (`Done t) c.esys
      )
      ++ (fun trans ->
	    cur_trans := Some trans;
	    Uq_engines.meta_engine
	      (f trans)
	    ++ (function
		  | `Done() -> 
		      eps_e (`Done ()) c.esys
		  | `Error (Plasma_error `enoient) ->
		      (** [ENOIENT] is perfectly legal - the inode
			  has been deleted in the meantime.
		       *)
		      eps_e (`Done ()) c.esys
		  | `Error (Plasma_error `econflict) ->
		      (** [ECONFLICT] means that another transaction locks
			  the inode. As this is outside the control of
			  the user, we handle it here automatically by
			  retrying the write.
		       *)
		      let t = pause.(min (Array.length pause - 1) rep) in
		      Netlog.logf `Debug
			"Cannot flush because of ECONFLICT; \
                         retrying in %f seconds"
			t;
		      (** Commit what we've written to prevent deadlocks: *)
		      commit_e trans 
		      ++ (fun () ->
			    cur_trans := None;
			    Plasma_util.delay_engine t 
			      (fun () -> 
				 trans_repeat mc (rep+1) f
			      )
			      c.esys
			 )
		  | other ->
		      eps_e (other :> _ Uq_engines.engine_state) c.esys
	       )
	 )
    )
  in

  let rec allocate_and_write mc rangelist =
    match rangelist with
      | (imin,imax) :: rangelist' ->
	  Netlog.logf `Debug
	    "Plasma_client.allocate_and_write: inode=%Ld imin=%Ld imax=%Ld"
	    inode imin imax;
	  trans_repeat mc 0
	    (fun trans ->
	       let ilen = Int64.succ(Int64.sub imax imin) in
	       ( Plasma_util.rpc_engine mc 
		   Fsys.V1.allocate_blocks'async
		   (trans.tid,inode,imin,ilen,true,Array.of_list c.pref_nodes)
		 >> fextract
	       )++ (fun blocklist ->
		      assert(blocklist <> [| |]);
		      let seqno = blocklist.(0).inode_seqno in
		      wrbuffers_e 
			c inodemod imin imax blocklist seqno n_success
		      ++ (fun () -> 
			    allocate_and_write mc rangelist'
			 )
		   )
	    )
      | [] ->
	  eps_e (`Done ()) c.esys 
  in

  let update_ii mc =
    Netlog.logf `Debug
      "Plasma_client.update_ii: inode=%Ld" inode;
    trans_repeat mc 0
      (fun trans ->
	 get_inodeinfo_e trans inode
	 ++ (fun ii ->
	       let ii' =
		 { ii with
		     eof = max ii.eof min_eof;
		 } in
	       set_inodeinfo_e trans inode ii'
	    )
      )
  in

  match inodemod.mod_err with
    | None ->
	Netlog.logf `Debug
	  "Plasma_client.flinode_e: starting \
           inode=%Ld old_fbuffers=%d n=%d"
	  inode c.f_buffers n;

	with_nn_session_e c
	  (fun mc ->
	     c.f_buffers <- c.f_buffers + n;
	     Uq_engines.meta_engine
	       (allocate_and_write mc !rangelist
		++ (fun () -> update_ii mc)
	       )
	     ++ (fun st ->
		   c.f_buffers <- c.f_buffers - n + !n_success;
		   (* add n_success because this has already been subtracted *)
		   Netlog.logf `Debug
		     "Plasma_client.flinode_e: done writing \
                      inode=%Ld new_fbuffers=%d success=%b"
		     inode c.f_buffers (st = `Done());
		   let st' = (st :> _ Uq_engines.engine_state) in
		   match st with
		     | `Done () ->
			 ( match !cur_trans with
			     | Some t -> commit_e t
			     | None -> eps_e (`Done ()) c.esys
			 )
		     | `Error e ->
			 Netlog.logf `Err
			   "Stopping flush after unexpected error %s"
			   (Netexn.to_string e);
			 inodemod.mod_err <- Some e;
			 ( match !cur_trans with
			     | Some t -> abort_e t
			     | None -> eps_e (`Done ()) c.esys
			 ) ++ (fun () -> eps_e st' c.esys)
		     | `Aborted ->
			 ( match !cur_trans with
			     | Some t -> abort_e t
			     | None -> eps_e (`Done ()) c.esys
			 ) ++ (fun () -> eps_e `Aborted c.esys)
		)
	  )       
    | Some err ->
	Netlog.logf `Debug
	  "Plasma_client.flinode_e: NOT starting because mod_err is set \
           inode=%Ld old_fbuffers=%d n=%d"
	  inode c.f_buffers n;
	eps_e (`Error err) c.esys


let set_dirty c b min_eof =
  assert(b.buf_valid);
  if b.buf_clean then (
    b.buf_clean <- false;
    remove_from_cbuffers c b;
    (** Add to inodemod *)
    let inodemod =
      try
	Hashtbl.find c.inodemods b.buf_inode
      with
	| Not_found ->
	    let m =
	      { mod_inode = b.buf_inode;
		mod_bufs = Hashtbl.create 5;
		mod_min_eof = 0L;
		mod_flush = (fun () -> assert false);
		mod_flush_ser = Uq_engines.serializer c.esys;
		mod_err = None
	      } in
	    (** Set the flush engine: *)
	    m.mod_flush <- (fun () -> flinode_e c m);
	    m in
    Hashtbl.replace c.inodemods b.buf_inode inodemod;
    Hashtbl.replace inodemod.mod_bufs b.buf_index b;
    inodemod.mod_min_eof <- max inodemod.mod_min_eof min_eof;
  );
  (** Add b to dbuffers or oldify *)
  BLRU.add b c.dbuffers;
  (** maybe we have to flush: *)
  maybe_flush_buffer c


let rec drop_inode c inode =
  try
    let im = Hashtbl.find c.inodemods inode in
    let _e =
      im.mod_flush_ser # serialized
	(fun _ ->
	   (** At some other place we could already have removed/replaced im, so
	       check for this
	    *)
	   let bufs = ref [] in
	   ( try
	       let im' = Hashtbl.find c.inodemods inode in
	       if im == im' then (
		 Hashtbl.iter
		   (fun index b ->
		      let b' = drop_buffer c b in
		      bufs := b' :: !bufs
		   )
		   im.mod_bufs;
		 Hashtbl.remove c.inodemods inode
	       )
	       else drop_inode c inode
	     with Not_found -> ()
	   );
	   (** Maybe somebody waits for buffers *)
	   List.iter
	     (fun b ->
		post_action c b
	     )
	     !bufs;
	   eps_e (`Done ()) c.esys
	) in
    ()
  with
    | Not_found -> ()


let reset_inode_error c inode =
  try
    let im = Hashtbl.find c.inodemods inode in
    im.mod_err <- None;
  with Not_found -> ()

		       
let write_e c inode pos sm spos len =
  let l = sm_length sm in
  if spos < 0 || len < 0 || spos > l-len then
    invalid_arg "Plasma_client.write_e";
  if pos < 0L || pos > Int64.sub Int64.max_int (Int64.of_int len) then
    invalid_arg "Plasma_client.write_e";

  let modify b bpos len'' min_eof =
    blit_to_buffer c sm spos b bpos len'';
    set_dirty c b min_eof;
  in

  let read_before_write mc blocksize inode index action =
    (** Start a transaction: *)
    with_new_nn_transaction_e c mc (fun x -> x)
      (fun mc trans ->
	 get_inodeinfo_e trans inode
	 ++ (fun ii ->
	       rdblock_e 
		 c mc (Some trans.tid) ii blocksize inode index action
	    )
      )
  in

  (** maybe we have an error to report *)
  let err_opt =
    try
      (Hashtbl.find c.inodemods inode).mod_err
    with Not_found -> None in

  match err_opt with
    | None ->

	with_nn_session_e c
	  (fun mc ->
	     let blocksize = c.blocksize in
	     let index =
	       Int64.div pos (Int64.of_int blocksize) in
	     let bpos = 
	       Int64.to_int
		 (Int64.sub pos
		    (Int64.mul index (Int64.of_int blocksize))) in
	     let len' =
	       min len (blocksize - bpos) in
	     
	     let min_eof =
	       Int64.add pos (Int64.of_int len') in
	     
	     if len' = 0 then
	       eps_e (`Done 0) c.esys
	     else (
	       let rec find_and_write () =
		 (** If the block is already known: Just write. *)
		 try
		   let b = Hashtbl.find c.buffers (inode,index) in
		   if not b.buf_valid then raise Not_found;
		   modify b bpos len' min_eof;
		   eps_e (`Done len') c.esys
		 with
		   | Not_found ->
		       (** When writing a full block do not
			   read the old version first
			*)
		       if len' = blocksize then (
			 find_buffer_e c blocksize inode index
			 ++ (fun b_opt ->
			       match b_opt with
				 | None -> find_and_write()
				 | Some b ->
				     (** The following is a temporary lie.
					 However, these settings trigger the
					 right things.
				      *)
				     b.buf_valid <- true;
				     b.buf_clean <- true;
				     modify b bpos len' min_eof;
				     eps_e (`Done len') c.esys
			    )
		       )
		       else (
			 (** Now we have to read the block first. We have to
			     do this in a new transaction.
			  *) 
			 read_before_write mc blocksize inode index
			   (fun b ->
			      modify b bpos len' min_eof
			   )
			 ++ (fun () -> eps_e (`Done len') c.esys)
		       )
	       in
	       find_and_write()
	     )
	  )
    | Some err ->
	(** Report the error and reset the error flag *)
	reset_inode_error c inode;
	eps_e (`Error err) c.esys


let write c inode pos s spos len = sync (write_e c inode pos s spos) len

let flush_e c inode pos len =
  if len < 0L then invalid_arg "Plasma_client.flush_e";
  (* pos,len: ignored in this implementation *)

  (** maybe we have an error to report *)
  let err_opt =
    try
      (Hashtbl.find c.inodemods inode).mod_err
    with Not_found -> None in

  match err_opt with
    | None ->
	let e = flush_inode_e c inode in
	Uq_engines.when_state
	  ~is_error:(fun _ -> reset_inode_error c inode)
	  e;
	e
    | Some err ->
	(** Report the error and reset the error flag *)
	reset_inode_error c inode;
	eps_e (`Error err) c.esys



let flush c inode pos len = sync (flush_e c inode pos) len


let flush_all_e c =
  let engines = ref [] in
  Hashtbl.iter
    (fun inode _ ->
       engines := (flush_inode_e c inode) :: !engines
    )
    c.inodemods;
  Plasma_util.msync_engine !engines c.esys

let flush_all c = sync flush_all_e c


let split_by f l =
  let ht = Hashtbl.create 7 in
  List.iter
    (fun x ->
       let k = f x in
       try
	 let p = Hashtbl.find ht k in
	 Hashtbl.replace ht k (x::p)
       with
	 | Not_found ->
	 Hashtbl.replace ht k [x]
    )
    l;
  let v =
    Hashtbl.fold
      (fun _ p acc -> List.rev p :: acc)
      ht
      [] in
  let kl =
    Hashtbl.fold
      (fun k _ acc -> k :: acc)
      ht
      [] in
  (v, kl)


type dn_shm =
    { shm_serial : int;
      shm_opath : string;
      mutable shm_blocks : int;
      mutable shm_free : (Netsys_mem.memory * int) list;
      shm_id : Plasma_shm.shm_id
    }

type supp_entry =
    [ `Supported of dn_shm
    | `Unsupported
    | `Requesting of unit Uq_engines.engine
    ]

type dn_shm_support = (string * int, supp_entry) Hashtbl.t
    (** Remembers for which (host,port) pair shared memory is supported *)


let get_shm_buf c shm blocksize =
  match shm.shm_free with
    | (m,idx) :: free' ->
	shm.shm_free <- free';
	m, idx
    | [] ->
	let fd = Plasma_util.open_shm_fd shm.shm_opath in
	let idx = shm.shm_blocks in
	shm.shm_blocks <- idx + 1;
	let pos = Int64.of_int (idx * blocksize) in
	let eof = (idx+1) * blocksize in
	Plasma_util.enlarge_shm fd eof;
	c.shm_mng # book_allocation shm.shm_id eof;
	let m = Plasma_util.map_shm fd pos blocksize in
	Unix.close fd;
	m, idx
    

let release_shm_buf shm m idx =
  shm.shm_free <- (m,idx) :: shm.shm_free


let destroy_shm c shm =
  (** The shm file is deleted by the datanode server when we close the
      TCP connection. For releasing memory, we also should unmap memory
      as quickly as possible.
   *)
  List.iter
    (fun (m,_) ->
       Netsys_mem.memory_unmap_file m
    )
    shm.shm_free;
  shm.shm_free <- [];
  ( try
      let fd = Plasma_util.open_shm_fd shm.shm_opath in
      Unix.ftruncate fd 0;
      Unix.close fd
    with
      | Unix.Unix_error(Unix.ENOENT,_,_) -> ()
  );
  c.shm_mng # book_allocation shm.shm_id 0


let rec request_shm_e c dc (supp : dn_shm_support) (h,p) =
  (** Check whether there is shm support, and return [dn_shm option] *)
  let dc_serial = Rpc_proxy.ManagedClient.mclient_serial dc in
  let hp = (h,p) in
  try
    let shm_entry = Hashtbl.find supp hp in
    match shm_entry with
      | `Unsupported -> 
	  eps_e (`Done None) c.esys
      | `Supported shm ->
	  if shm.shm_serial <> dc_serial then (
	    destroy_shm c shm;
	    raise Not_found
	  );
	  eps_e (`Done(Some shm)) c.esys
      | `Requesting e ->
	  e ++ (fun () -> request_shm_e c dc supp hp)
  with
    | Not_found ->
	Netlog.logf `Debug
	  "Plasma_client: Checking SHM support for %s:%d" h p;
	let e =
	  Util.rpc_engine dc Dn.V1.alloc_shm_if_local'async
	    () 
	  ++ (fun path_opt ->
		let shm_opt =
		  match path_opt with
		    | None -> 
			Netlog.logf `Debug
			  "Plasma_client: No SHM support for %s:%d" h p;
			Hashtbl.replace supp hp `Unsupported;
			None
		    | Some path -> 
			Netlog.logf `Debug
			  "Plasma_client: Got SHM support for %s:%d with %s"
			  h p path;
			let shm =
			  { shm_serial = dc_serial;
			    shm_opath = path;
			    shm_blocks = 0;
			    shm_free = [];
			    shm_id = c.shm_mng # new_id();
			  } in
			Hashtbl.replace supp hp (`Supported shm);
			Some shm in
		eps_e (`Done shm_opt) c.esys
	     ) in
	let e' = e ++ (fun _ -> eps_e (`Done ()) c.esys) in
	Hashtbl.replace supp hp (`Requesting e');
	e


let copy_in_e c inode pos fd len topo =
  (* TODO: check pos, len *)

  if pos < 0L || len < 0L then invalid_arg "Plasma_client.copy_in_e";

  let get_info_e mc tid =
    (** Get information: inodeinfo *)
    (Util.rpc_engine mc Fsys.V1.get_inodeinfo'async (tid,inode))
    >> fextract in

  let dn_shm = Hashtbl.create 5 in

  let data_q = Queue.create() in
  let get_data_buf blocksize =
    try
      Queue.take data_q
    with
      | Queue.Empty ->
	  Netsys_mem.alloc_memory_pages blocksize in

  let write_block_e blocksize (block, block_fd_pos, block_len) =
    let read_data data =
      ignore(Unix.LargeFile.lseek fd block_fd_pos Unix.SEEK_SET);
      let p = Netsys_mem.mem_read fd data 0 (Int64.to_int block_len) in
      if (p < Int64.to_int block_len) then (
	Netlog.logf `Warning
	  "copy_in: padding with zeros inode=%Ld fd_pos=%Ld block_len=%Ld p=%d"
	  inode block_fd_pos block_len p;
	Bigarray.Array1.fill
	  (Bigarray.Array1.sub data p (blocksize-p)) '\000'
      )
    in

    Netlog.logf `Debug
      "write_block: node=%s index=%Ld" block.node block.index;
    try
      let hp = Util.parse_host_port block.node in
      dn_session_e c hp
      ++ (fun dc ->
	    (** Check whether there is shm support *)
	    request_shm_e c dc dn_shm hp ++
	      (fun shm_opt ->
		 let wreq, release_buf =
		   match shm_opt with
		     | None ->
			 let data = get_data_buf blocksize in
			 read_data data;
			 let ms =
			   Xdr_mstring.memory_based_mstrings # create_from_memory
			     data
			     0
			     blocksize
			     false in
			 let release_buf() =
			   Queue.add data data_q in
			(`dnch_rpc ms, release_buf)
		     | Some shm ->
			 let m, idx = get_shm_buf c shm blocksize in
			 read_data m;
			 let release_buf() =
			   release_shm_buf shm m idx in
			 (`dnch_shm { shm_path = shm.shm_opath;
				      shm_offset = Int64.of_int (idx*blocksize);
				      shm_length = blocksize
				    },
			  release_buf) in
		 let e =
		   Util.rpc_engine dc Dn.V1.write'async
		     (block.block, wreq, 
		      block.safetrans_id, block.safetrans_vfy) in
		 Uq_engines.when_state
		  ~is_done:(fun _ -> release_buf())
		   e;
		 e
	      )
	 )
    with
      | error ->
	  Netlog.logf `Err
	    "Exception reading from file: %s"
	    (Netexn.to_string error);
	  eps_e (`Error error) c.esys
  in

  let copy_block_e blocksize orig_block dest_block =
    Netlog.logf `Debug
      "copy_block: from_node=%s from_index=%Ld to_node=%s to_index=%Ld"
      orig_block.node orig_block.index dest_block.node dest_block.index;
    try
      let hp = Util.parse_host_port orig_block.node in
      dn_session_e c hp
      ++ (fun dc ->
	    Util.rpc_engine dc Dn.V1.copy'async
	      (orig_block.block,
	       dest_block.node, dest_block.identity, dest_block.block,
	       orig_block.safetrans_id, orig_block.safetrans_vfy,
	       dest_block.safetrans_id, dest_block.safetrans_vfy
	      )
	 )
    with
      | error ->
	  Netlog.logf `Err
	    "Exception in copy_block_e: %s"
	    (Netexn.to_string error);
	  eps_e (`Error error) c.esys
  in

  let dn_parallelism blocksize =
    (** Write this many blocks to a datanode in parallel.
	
	We usually want 8 blocks here, but we also want to limit the memory
	if the blocksize is high. So try to limit to 64M
     *)
    let shm_cond = c.shm_mng # shm_condition in
    let lim =
      match shm_cond with
	| `Normal -> 64 * 1024 * 1024 
	| `Tight -> 16 * 1024 * 1024 in
    let lim_p = max 1 (lim / blocksize) in
    match shm_cond with
      | `Normal -> min 8 lim_p
      | `Tight -> min 4 lim_p in

  let rec fadvise f ext_blocklist n =
    if n = 0 then
      ()
    else (
      match ext_blocklist with
	| x :: ext_blocklist' ->
	    let (pos,len) = f x in
	    Netsys_posix.fadvise fd pos len Netsys_posix.FADV_WILLNEED;
	    fadvise f ext_blocklist' (n-1)
	| [] -> ()
    )
  in

  let star_write_stream_e blocksize ext_blocklist =
    let n = ref 0 in
    let ext_blocklist = ref ext_blocklist in
    let dn_par = dn_parallelism blocksize in

    let stream_e, signal =
      Plasma_util.signal_engine c.esys in

    let rec maybe_add_more () =
      if Netsys_posix.have_fadvise () then
	fadvise (fun (_,pos,len) -> (pos,len)) !ext_blocklist 4;
      if !n < dn_par then (
	match !ext_blocklist with
	  | (block, block_fd_pos, block_len) :: l ->
	      ext_blocklist := l;
	      let _e =
		write_block_e blocksize (block, block_fd_pos, block_len)
		++ (fun () ->
		      decr n;
		      maybe_add_more();
		      eps_e (`Done ()) c.esys
		   ) in
	      incr n;
	      maybe_add_more()
	  | [] ->
	      ()
      );
      if !n=0 then
	signal(`Done())
    in

    maybe_add_more();
    stream_e
  in

  let datanodes = Hashtbl.create 7 in

  let sync_datanode node =
    let hp = Util.parse_host_port node in
    dn_session_e c hp
    ++ (fun dc ->
	  Util.rpc_engine dc Dn.V1.sync'async ()   (* :-) *)
       )
  in

  let star_write_blocks_e blocksize ext_blocklist =
    (** Split the blocklist by datanode, and write for each datanode
	sequentially. 
     *)
    let ext_blocklist_by_node, nodes =
      split_by (fun (block,_,_) -> block.node) ext_blocklist in
    List.iter (fun n -> Hashtbl.replace datanodes n ()) nodes;
    Plasma_util.msync_engine
      (List.map
	 (fun ext_blocklist' ->
	    star_write_stream_e blocksize ext_blocklist'
	 )
	 ext_blocklist_by_node
      )
      c.esys
    ++ (fun () ->
	  Plasma_util.msync_engine
	    (List.map
	       (fun node ->
		  sync_datanode node
	       )
	       nodes
	    )
	    c.esys
       )
  in

  let rec chain_copy_stream_e blocksize orig_block ext_blocklist =
    match ext_blocklist with
      | dest_block :: l ->
	  copy_block_e blocksize orig_block dest_block
	  ++ (fun () ->
		chain_copy_stream_e blocksize dest_block l
	     )
      | [] ->
	  eps_e (`Done ()) c.esys
  in

  let chain_write_stream_e blocksize ext_blocklist_by_block =
    let n = ref 0 in
    let ext_blocklist_by_block = ref ext_blocklist_by_block in
    let dn_par = dn_parallelism blocksize in

    let stream_e, signal =
      Plasma_util.signal_engine c.esys in

    let g = Unixqueue.new_group c.esys in

    let rec maybe_add_more () =
      if Netsys_posix.have_fadvise () then
	fadvise
	  (fun l ->
	     match l with
	       | (_,pos,len) :: _ -> (pos,len)
	       | _ -> assert false
	  )
	  !ext_blocklist_by_block 4;
      if !n < dn_par then (
	match !ext_blocklist_by_block with
	  | ext_blocklist :: l ->
	      ext_blocklist_by_block := l;

	      ( match ext_blocklist with
		  | (block, block_fd_pos, block_len) :: xl ->

		      let _e =
			write_block_e blocksize (block, block_fd_pos, block_len)
			++ (fun () ->
			      chain_copy_stream_e 
				blocksize block 
				(List.map (fun (b,_,_) -> b) xl)
			   ) 
			++ (fun () ->
			      decr n;
			      maybe_add_more();
			      eps_e (`Done ()) c.esys
			   ) in
		      incr n;
		      (* This ensures that the RPCs are submitted in the
			 right order:
		       *)
		      Unixqueue.once c.esys g 0.0 maybe_add_more
		  | [] ->
		      assert false
	      )
	  | [] ->
	      ()
      );
      if !n=0 then
	signal(`Done())
    in

    maybe_add_more();
    stream_e
  in

  let chain_write_blocks_e blocksize ext_blocklist =
    (** Split the blocklist by blocks
     *)
    let ext_blocklist_by_block0, _ =
      split_by (fun (block,_,_) -> block.index) ext_blocklist in
    let _, nodes =
      split_by (fun (block,_,_) -> block.node) ext_blocklist in

    let ext_blocklist_by_block =
      List.sort
	(fun bl1 bl2 ->
	   match bl1, bl2 with
	     | ((b1,_,_) :: _), ((b2,_,_) :: _) -> compare b1.block b2.block
	     | _ -> assert false
	)
	ext_blocklist_by_block0 in

    List.iter (fun n -> Hashtbl.replace datanodes n ()) nodes;
    chain_write_stream_e blocksize ext_blocklist_by_block
    ++ (fun () ->
	  Plasma_util.msync_engine
	    (List.map
	       (fun node ->
		  sync_datanode node
	       )
	       nodes
	    )
	    c.esys
       )
  in

  let write_blocks_e =
    match topo with
      | `Star -> star_write_blocks_e
      | `Chain -> chain_write_blocks_e in

(*
  let sync_datanodes() =
    Plasma_util.msync_engine
      (List.map
	 (fun node ->
	    sync_datanode node
	 )
	 (Hashtbl.fold (fun n _ acc -> n::acc) datanodes [])
      )
      c.esys
  in
 *)

  let upload_e mc ii blocksize fd_pos pos len =
    (** Uploads data from [fd_pos] to [fd_pos+len] in a transaction.
        This includes block allocation, and writing data blocks, and
        updating EOF.
     *)
    assert(Int64.rem pos (Int64.of_int blocksize) = 0L);
    let blocks =
      Int64.succ (Int64.div (Int64.pred len) (Int64.of_int blocksize)) in
    let idx =
      Int64.div pos (Int64.of_int blocksize) in
    let eof =
      max ii.eof (Int64.add pos len) in
    with_new_nn_transaction_e c mc no_extract
      (fun mc trans ->
	 ( Util.rpc_engine 
	     mc Fsys.V1.allocate_blocks'async(trans.tid,inode,idx,blocks,true,
					      Array.of_list c.pref_nodes
					     )
	   >> fextract
	 ) 
	 ++ (fun blocklist ->
	       let ext_blocklist =
		 Array.map
		   (fun block ->
		      let rel_pos =
			Int64.mul 
			  (Int64.sub block.index idx) 
			  (Int64.of_int blocksize) in
		      let block_fd_pos =
			Int64.add fd_pos rel_pos in
		      let block_len =
			min 
			  (Int64.sub len rel_pos)
			  (Int64.of_int blocksize) in
		      (block, block_fd_pos, block_len)
		   )
		   blocklist in
	       write_blocks_e blocksize (Array.to_list ext_blocklist)
	       ++ (fun _ ->
		     if eof = ii.eof then
		       new Uq_engines.epsilon_engine (`Done ()) c.esys
		     else
		       Util.rpc_engine 
			 mc
			 Fsys.V1.update_inodeinfo'async
			 (trans.tid, inode, { ii with eof = eof })
		       >> fextract_unit
		  )
	    )
      ) in

  (* TODO: upload_attempts_e = upload_e with restart on error *)

  let rec upload_rest_e mc ii blocksize fd_pos pos len =
    (** Uploads the remaining data, starting at pos/fd_pos.
	Finally sync each datanode.
     *)
    let blocksizeL = Int64.of_int blocksize in
    let len_limit = 
      (** Limit to 1G by now. Should be more intelligent *)
      max blocksizeL 1073741824L in
    let eff_len =
      min len len_limit in
    if eff_len > 0L then
      upload_e mc ii blocksize fd_pos pos eff_len
      ++ (fun _ ->
	    let fd_pos' = Int64.add fd_pos eff_len in
	    let pos' = Int64.add pos eff_len in
	    let len' = Int64.sub len eff_len in
	    upload_rest_e mc ii blocksize fd_pos' pos' len'
	 )
    else (
      Hashtbl.iter (fun _ e -> 
		      match e with
			| `Supported shm -> destroy_shm c shm
			| _ -> ()
		   ) dn_shm;
      eps_e (`Done()) c.esys
    ) in
  
  let fd_pos =
    Unix.LargeFile.lseek fd 0L Unix.SEEK_CUR in
  
  with_nn_session_e c
    (fun mc ->
       with_new_nn_transaction_e c mc no_extract
	 (fun mc trans ->
	    get_info_e mc trans.tid
	    ++ (fun ii ->
		  let blocksize = c.blocksize in
		  upload_rest_e mc ii blocksize fd_pos pos len
	       )
	 )
    )

let copy_in c inode pos fd len topo = sync (copy_in_e c inode pos fd len) topo


let copy_out_e c inode pos fd len =
  (* TODO: check pos, len *)

  if pos < 0L || len < 0L then invalid_arg "Plasma_client.copy_out_e";

  let get_info_e mc tid =
    (** Get information: inodeinfo *)
    Util.rpc_engine mc Fsys.V1.get_inodeinfo'async (tid,inode)
    >> fextract
  in
  
  let dn_shm = Hashtbl.create 5 in

  let read_block_e blocksize (block, block_fd_pos, block_len) =
    Netlog.logf `Debug
      "read_block: node=%s index=%Ld" block.node block.index;
    let hp = Util.parse_host_port block.node in
    let blen = Int64.to_int block_len in
    dn_session_e c hp
    ++ (fun dc ->
	  (** Check whether there is shm support *)
	  request_shm_e c dc dn_shm hp ++
	    (fun shm_opt ->
	       let rreq, shm_details =
		 match shm_opt with
		   | None ->
		       (`dnch_rpc, None)
		   | Some shm ->
		       let m, idx = get_shm_buf c shm blocksize in
		       (`dnch_shm { shm_path = shm.shm_opath;
				    shm_offset = Int64.of_int (idx*blocksize);
				    shm_length = blocksize
				  },
			Some(shm,m,idx)
		       ) in
	       Util.rpc_engine 
		 dc Dn.V1.read'async 
		 (rreq, block.block, 0, blen, 
		  block.safetrans_id, block.safetrans_vfy)
	       ++ (fun data_designator ->
		     try
		       ignore(Unix.LargeFile.lseek 
				fd block_fd_pos Unix.SEEK_SET);
		       ( match data_designator, shm_details with
			   | `dnch_rpc data, None ->
			       let data_s, offs = data # as_string in
			       Netsys.really_gwrite
				 `Read_write fd data_s offs blen;
			   | `dnch_shm, Some(shm,m,idx) ->
			       ignore(Netsys_mem.mem_write fd m 0 blen);
			       release_shm_buf shm m idx
			   | _ ->
			       assert false
		       );
		       eps_e (`Done ()) c.esys
		     with
		       | error ->
			   Netlog.logf `Err
			     "Exception writing to file: %s"
			     (Netexn.to_string error);
			   eps_e (`Error error) c.esys
		  )
	    )
       )
  in
  
  (** Reading in optimal order is complicated:
      - each block is on several datanodes, and we can pick one
       - the datanodes may be loaded differently, so the read speed may
         also differ
       - we can request several blocks from each datanode in parallel

      Solution:
       - for each datanode we have a separate stream of blocks that are
         being read
       - for each stream we request N blocks at the same time. If a
         block is returned to us, we request another one

   *)

  let requested = Hashtbl.create 113 in
  (** If a block is requested from a datanode, the block is put into this
      hash table, so that the streams for other datanodes can skip this
   *)

  let dn_parallelism blocksize =
    (** Read this many blocks to a datanode in parallel.
	
	We usually want 8 blocks here, but we also want to limit the memory
	if the blocksize is high. So try to limit to 64M
     *)
    let shm_cond = c.shm_mng # shm_condition in
    let lim =
      match shm_cond with
	| `Normal -> 64 * 1024 * 1024 
	| `Tight -> 16 * 1024 * 1024 in
    let lim_p = max 1 (lim / blocksize) in
    match shm_cond with
      | `Normal -> min 8 lim_p
      | `Tight -> min 4 lim_p in

  let read_stream_e blocksize ext_blocklist =
    (** Reads the blocks which are all from the same datanode. Skip blocks
	already requested
     *)
    let n = ref 0 in
    let ext_blocklist = ref ext_blocklist in
    let dn_par = dn_parallelism blocksize in

    let stream_e, signal =
      Plasma_util.signal_engine c.esys in

    let rec maybe_add_more () =
      if !n < dn_par then (
	match !ext_blocklist with
	  | (block, block_fd_pos, block_len) :: l ->
	      ext_blocklist := l;
	      if Hashtbl.mem requested block.block then
		maybe_add_more()
	      else (
		Hashtbl.add requested block.block ();
		let _e =
		  read_block_e blocksize (block, block_fd_pos, block_len)
		  ++ (fun () ->
			decr n;
			maybe_add_more();
			eps_e (`Done ()) c.esys
		     ) in
		incr n;
		maybe_add_more()
	      )
	  | [] ->
	      ()
      );
      if !n=0 then
	signal(`Done())
    in

    maybe_add_more();
    stream_e
  in

  let read_blocks_e blocksize ext_blocklist =
    (** Split the blocklist by datanode, and read for each datanode
	sequentially. 
     *)
    Hashtbl.clear requested;
    let ext_blocklist_by_node, nodes =
      split_by (fun (block,_,_) -> block.node) ext_blocklist in
    Plasma_util.msync_engine
      (List.map
	 (fun ext_blocklist' ->
	    read_stream_e blocksize ext_blocklist'
	 )
	 ext_blocklist_by_node
      )
      c.esys
  in

  let download_e mc tid ii blocksize fd_pos pos len =
    (** Downloads data from [fd_pos] to [fd_pos+len] *)
    let blocks =
      Int64.succ (Int64.div (Int64.pred len) (Int64.of_int blocksize)) in
    let idx =
      Int64.div pos (Int64.of_int blocksize) in
    ( Util.rpc_engine mc Fsys.V1.get_blocks'async(tid,inode,idx,blocks)
      >> fextract
    ) 
    ++ (fun blocklist ->
	  (** We only filter here dead nodes out, and check that
	      each block is available on at least one node.
	   *)
	  let block_ht = Hashtbl.create 7 in
	  Array.iter
	    (fun block ->
	       Hashtbl.replace block_ht block.index false
	    )
	    blocklist;
	  
	  let ext_blocklist =
	    List.map
	      (fun block ->
		 Hashtbl.replace block_ht block.index true;
		 let rel_pos =
		   Int64.mul 
		     (Int64.sub block.index idx) 
		     (Int64.of_int blocksize) in
		 let block_fd_pos =
		   Int64.add fd_pos rel_pos in
		 let block_len =
		   min (Int64.sub len rel_pos) (Int64.of_int blocksize) in
		 (block, block_fd_pos, block_len)
	      )
	      (List.filter
		 (fun block ->
		    block.node_alive)
		 (Array.to_list blocklist)) in
	  
	  let all_ok =
	    Hashtbl.fold (fun _ ok acc -> ok && acc) block_ht true in
	  
	  if all_ok then
	    read_blocks_e blocksize ext_blocklist
	  else
	    plasma_error_e c `eio
       )
  in

  let rec download_rest_e mc tid ii blocksize fd_pos pos len =
    (** Downloads the remaining data, starting at pos/fd_pos *)
    let blocksizeL = Int64.of_int blocksize in
    let len_limit = 
      (** Limit to 1G by now. Should be more intelligent *)
      max blocksizeL 1073741824L in
    let eff_len =
      min len len_limit in
    new Uq_engines.seq_engine
      (download_e mc tid ii blocksize fd_pos pos eff_len)
      (fun _ ->
	 if eff_len < len then
	   let fd_pos' = Int64.add fd_pos eff_len in
	   let pos' = Int64.add pos eff_len in
	   let len' = Int64.sub len eff_len in
	   download_rest_e mc tid ii blocksize fd_pos' pos' len'
	 else (
	   Hashtbl.iter (fun _ e -> 
			   match e with
			     | `Supported shm -> destroy_shm c shm
			     | _ -> ()
			) dn_shm;
	   new Uq_engines.epsilon_engine (`Done ()) c.esys
	 )
      ) in

  let fd_pos =
    Unix.LargeFile.lseek fd 0L Unix.SEEK_CUR in

  let fd_eof =
    Unix.LargeFile.lseek fd 0L Unix.SEEK_END in

  if fd_pos > Int64.sub fd_eof len then (
    let new_eof = Int64.add fd_pos len in
    Netlog.logf `Debug "copy_out (pid %d): truncate fd=%Ld new_eof=%Ld"
      (Unix.getpid()) (Netsys.int64_of_file_descr fd) new_eof;
    Unix.LargeFile.ftruncate fd new_eof
  );
  
  with_nn_session_e c
    (fun mc ->
       with_new_nn_transaction_e c mc no_extract
	 (fun mc trans ->
	    get_info_e mc trans.tid
	    ++ (fun ii ->
		  let blocksize = c.blocksize in
		  download_rest_e mc trans.tid ii blocksize fd_pos pos len
	       )
	 )
    )

let copy_out c inode pos fd len = sync (copy_out_e c inode pos fd) len


let rec retry name n f arg =
  try
    let r = f arg in
    r
  with
    | Plasma_error code as error ->
	Netlog.logf `Err
	  "Plasma error for %s: %s"
	  name
	  (Plasma_util.string_of_errno code);
	next_attempt name n f arg error
    | Rpc_client.Message_timeout as error ->
	Netlog.logf `Err
	  "Plasma error for %s: Timeout"
	  name;
	next_attempt name n f arg error
    | error ->
	Netlog.logf `Crit
	  "Exception accessing %s: %s"
	  name
	  (Netexn.to_string error);
	raise error

and next_attempt name n f arg error =
  if n > 0 then (
    Unix.sleep 60;
    Netlog.logf `Err
      "Retrying operation after error";
    retry name (n-1) f arg
  ) else (
    Netlog.logf `Err "No retry";
    raise error
  )
  
